@@ -2,7 +2,7 @@
 
 #include "zstring.h"
 
-#ifdef __unix__
+#if defined(__unix__) || defined(__HAIKU__)
 FString GetUserFile (const char *path);
 #endif
 FString M_GetAppDataPath(bool create);
@@ -16,6 +16,7 @@ FString M_GetDemoPath();
 
 FString M_GetNormalizedPath(const char* path);
 
+
 #ifdef __APPLE__
 FString M_GetMacAppSupportPath(const bool create = true);
 void M_GetMacSearchDirectories(FString& user_docs, FString& user_app_support, FString& local_app_support);
