@@ -34,7 +34,7 @@
 */
 
 #include <assert.h>
-#include "templates.h"
+
 #include "g_level.h"
 #include "filesystem.h"
 #include "cmdlib.h"
@@ -50,6 +50,7 @@
 #include "g_levellocals.h"
 #include "events.h"
 #include "i_system.h"
+#include "screenjob.h"
 
 static TArray<cluster_info_t> wadclusterinfos;
 TArray<level_info_t> wadlevelinfos;
@@ -163,6 +164,20 @@ bool CheckWarpTransMap (FString &mapname, bool substitute)
 //
 //==========================================================================
 
+bool SecretLevelVisited()
+{
+	for (unsigned int i = 0; i < wadlevelinfos.Size(); i++)
+		if ((wadlevelinfos[i].flags3 & LEVEL3_SECRET) && (wadlevelinfos[i].flags & LEVEL_VISITED))
+			return true;
+
+	return false;
+}
+
+//==========================================================================
+//
+//
+//==========================================================================
+
 static int FindWadClusterInfo (int cluster)
 {
 	for (unsigned int i = 0; i < wadclusterinfos.Size(); i++)
@@ -249,10 +264,13 @@ void level_info_t::Reset()
 	else
 		flags2 = LEVEL2_LAXMONSTERACTIVATION;
 	flags3 = 0;
+	LightningSound = "world/thunder";
 	Music = "";
 	LevelName = "";
 	AuthorName = "";
+	MapLabel = "";
 	FadeTable = "COLORMAP";
+	CustomColorMap = "COLORMAP";
 	WallHorizLight = -8;
 	WallVertLight = +8;
 	F1Pic = "";
@@ -273,8 +291,12 @@ void level_info_t::Reset()
 	Translator = "";
 	RedirectType = NAME_None;
 	RedirectMapName = "";
+	RedirectCVAR = NAME_None;
+	RedirectCVARMapName = "";
 	EnterPic = "";
 	ExitPic = "";
+	EnterAnim = "";
+	ExitAnim = "";
 	Intermission = NAME_None;
 	deathsequence = NAME_None;
 	slideshow = NAME_None;
@@ -301,7 +323,8 @@ void level_info_t::Reset()
 	lightadditivesurfaces = -1;
 	skyrotatevector = FVector3(0, 0, 1);
 	skyrotatevector2 = FVector3(0, 0, 1);
-
+	lightblendmode = ELightBlendMode::DEFAULT;
+	tonemap = ETonemapMode::None;
 }
 
 
@@ -317,10 +340,10 @@ FString level_info_t::LookupLevelName(uint32_t *langtable)
 	if (flags & LEVEL_LOOKUPLEVELNAME)
 	{
 		const char *thename;
-		const char *lookedup = GStrings.GetString(LevelName, langtable);
+		const char *lookedup = GStrings.CheckString(LevelName.GetChars(), langtable);
 		if (lookedup == NULL)
 		{
-			thename = LevelName;
+			thename = LevelName.GetChars();
 		}
 		else
 		{
@@ -377,15 +400,38 @@ level_info_t *level_info_t::CheckLevelRedirect ()
 				if (playeringame[i] && players[i].mo->FindInventory(type))
 				{
 					// check for actual presence of the map.
-					if (P_CheckMapData(RedirectMapName))
+					if (P_CheckMapData(RedirectMapName.GetChars()))
 					{
-						return FindLevelInfo(RedirectMapName);
+						return FindLevelInfo(RedirectMapName.GetChars());
 					}
 					break;
 				}
 			}
 		}
 	}
+	if (RedirectCVAR != NAME_None)
+	{
+		auto var = FindCVar(RedirectCVAR.GetChars(), NULL);
+		if (var && (var->GetRealType() == CVAR_Bool) && !(var->GetFlags() & CVAR_IGNORE))	// only check Bool cvars that are currently defined
+		{
+			if (var->GetFlags() & CVAR_USERINFO)
+			{
+				// user sync'd cvar, check for all players
+				for (int i = 0; i < MAXPLAYERS; ++i)
+				{
+					if (playeringame[i] && (var = GetCVar(i, RedirectCVAR.GetChars())))
+					{
+						if (var->ToInt())
+							if (P_CheckMapData(RedirectCVARMapName.GetChars()))
+								return FindLevelInfo(RedirectCVARMapName.GetChars());
+					}
+				}
+			}
+			else if (var->ToInt())
+				if (P_CheckMapData(RedirectCVARMapName.GetChars()))
+					return FindLevelInfo(RedirectCVARMapName.GetChars());
+		}
+	}
 	return NULL;
 }
 
@@ -741,6 +787,30 @@ void FMapInfoParser::ParseMusic(FString &name, int &order)
 	}
 }
 
+//==========================================================================
+//
+//
+//
+//==========================================================================
+
+void FMapInfoParser::ParseCutscene(CutsceneDef& cdef, bool allow_function)
+{
+	FString sound;
+	sc.MustGetStringName("{");
+	while (!sc.CheckString("}"))
+	{
+		sc.MustGetString();
+		if (sc.Compare("video")) { ParseAssign(); sc.MustGetString(); cdef.video = sc.String; cdef.function = ""; }
+		else if (sc.Compare("function") && allow_function) { ParseAssign(); sc.SetCMode(false); sc.MustGetString(); sc.SetCMode(true); cdef.function = sc.String; cdef.video = ""; }
+		else if (sc.Compare("sound")) { ParseAssign(); sc.MustGetString(); cdef.soundName = sc.String; }
+		else if (sc.Compare("soundid")) { ParseAssign(); sc.MustGetNumber(); cdef.soundID = sc.Number; }
+		else if (sc.Compare("fps")) { ParseAssign();  sc.MustGetNumber();  cdef.framespersec = sc.Number; }
+		//else if (sc.Compare("transitiononly")) cdef.transitiononly = true;
+		else if (sc.Compare("delete")) { cdef.function = "none"; cdef.video = ""; } // this means 'play nothing', not 'not defined'.
+		else if (sc.Compare("clear")) cdef = {};
+	}
+}
+
 //==========================================================================
 //
 // ParseCluster
@@ -779,7 +849,7 @@ void FMapInfoParser::ParseCluster()
 			else
 			{
 				FStringf testlabel("CLUSTERENTER%d", clusterinfo->cluster);
-				if (GStrings.MatchDefaultString(testlabel, clusterinfo->EnterText))
+				if (GStrings.MatchDefaultString(testlabel.GetChars(), clusterinfo->EnterText.GetChars()))
 				{
 					clusterinfo->EnterText = testlabel;
 					clusterinfo->flags |= CLUSTER_LOOKUPENTERTEXT;
@@ -794,7 +864,7 @@ void FMapInfoParser::ParseCluster()
 			else
 			{
 				FStringf testlabel("CLUSTEREXIT%d", clusterinfo->cluster);
-				if (GStrings.MatchDefaultString(testlabel, clusterinfo->ExitText))
+				if (GStrings.MatchDefaultString(testlabel.GetChars(), clusterinfo->ExitText.GetChars()))
 				{
 					clusterinfo->ExitText = testlabel;
 					clusterinfo->flags |= CLUSTER_LOOKUPEXITTEXT;
@@ -845,6 +915,18 @@ void FMapInfoParser::ParseCluster()
 		{
 			clusterinfo->flags |= CLUSTER_EXITTEXTINLUMP;
 		}
+		else if (sc.Compare("intro"))
+		{
+			ParseCutscene(clusterinfo->intro);
+		}
+		else if (sc.Compare("outro"))
+		{
+			ParseCutscene(clusterinfo->outro);
+		}
+		else if (sc.Compare("gameover"))
+		{
+			ParseCutscene(clusterinfo->gameover);
+		}
 		else if (!ParseCloseBrace())
 		{
 			// Unknown
@@ -859,7 +941,7 @@ void FMapInfoParser::ParseCluster()
 	// Remap Hexen's CLUS?MSG lumps to the string table, if applicable. The code here only checks what can actually be in an IWAD.
 	if (clusterinfo->flags & CLUSTER_EXITTEXTINLUMP)
 	{
-		int lump = fileSystem.CheckNumForFullName(clusterinfo->ExitText, true);
+		int lump = fileSystem.CheckNumForFullName(clusterinfo->ExitText.GetChars(), true);
 		if (lump > 0)
 		{
 			// Check if this comes from either Hexen.wad or Hexdd.wad and if so, map to the string table.
@@ -868,7 +950,7 @@ void FMapInfoParser::ParseCluster()
 			if (fn && (!stricmp(fn, "HEXEN.WAD") || !stricmp(fn, "HEXDD.WAD")))
 			{
 				FStringf key("TXT_%.5s_%s", fn, clusterinfo->ExitText.GetChars());
-				if (GStrings.exists(key))
+				if (GStrings.exists(key.GetChars()))
 				{
 					clusterinfo->ExitText = key;
 					clusterinfo->flags &= ~CLUSTER_EXITTEXTINLUMP;
@@ -933,6 +1015,13 @@ DEFINE_MAP_OPTION(next, true)
 	parse.ParseNextMap(info->NextMap);
 }
 
+DEFINE_MAP_OPTION(lightningsound, true)
+{
+	parse.ParseAssign();
+	parse.sc.MustGetString();
+	info->LightningSound = parse.sc.String;
+}
+
 DEFINE_MAP_OPTION(author, true)
 {
 	parse.ParseAssign();
@@ -940,6 +1029,13 @@ DEFINE_MAP_OPTION(author, true)
 	info->AuthorName = parse.sc.String;
 }
 
+DEFINE_MAP_OPTION(label, true)
+{
+	parse.ParseAssign();
+	parse.sc.MustGetString();
+	info->MapLabel = parse.sc.String;
+}
+
 DEFINE_MAP_OPTION(secretnext, true)
 {
 	parse.ParseAssign();
@@ -1091,6 +1187,12 @@ DEFINE_MAP_OPTION(fadetable, true)
 	parse.ParseLumpOrTextureName(info->FadeTable);
 }
 
+DEFINE_MAP_OPTION(colormap, true)
+{
+	parse.ParseAssign();
+	parse.ParseLumpOrTextureName(info->CustomColorMap);
+}
+
 DEFINE_MAP_OPTION(evenlighting, true)
 {
 	info->WallVertLight = info->WallHorizLight = 0;
@@ -1138,6 +1240,11 @@ DEFINE_MAP_OPTION(gravity, true)
 	info->gravity = parse.sc.Float;
 }
 
+DEFINE_MAP_OPTION(nogravity, true)
+{
+	info->gravity = DBL_MAX;
+}
+
 DEFINE_MAP_OPTION(aircontrol, true)
 {
 	parse.ParseAssign();
@@ -1173,6 +1280,20 @@ DEFINE_MAP_OPTION(enterpic, true)
 	info->EnterPic = parse.sc.String;
 }
 
+DEFINE_MAP_OPTION(exitanim, true)
+{
+	parse.ParseAssign();
+	parse.sc.MustGetString();
+	info->ExitAnim = parse.sc.String;
+}
+
+DEFINE_MAP_OPTION(enteranim, true)
+{
+	parse.ParseAssign();
+	parse.sc.MustGetString();
+	info->EnterAnim = parse.sc.String;
+}
+
 DEFINE_MAP_OPTION(specialaction, true)
 {
 	parse.ParseAssign();
@@ -1205,8 +1326,8 @@ DEFINE_MAP_OPTION(PrecacheSounds, true)
 	do
 	{
 		parse.sc.MustGetString();
-		FSoundID snd = parse.sc.String;
-		if (snd == 0)
+		FSoundID snd = S_FindSound(parse.sc.String);
+		if (snd == NO_SOUND)
 		{
 			parse.sc.ScriptMessage("Unknown sound \"%s\"", parse.sc.String);
 		}
@@ -1261,6 +1382,15 @@ DEFINE_MAP_OPTION(redirect, true)
 	parse.ParseNextMap(info->RedirectMapName);
 }
 
+DEFINE_MAP_OPTION(cvar_redirect, true)
+{
+	parse.ParseAssign();
+	parse.sc.MustGetString();
+	info->RedirectCVAR = parse.sc.String;
+	parse.ParseComma();
+	parse.ParseNextMap(info->RedirectCVARMapName);
+}
+
 DEFINE_MAP_OPTION(sndseq, true)
 {
 	parse.ParseAssign();
@@ -1449,7 +1579,8 @@ DEFINE_MAP_OPTION(lightmode, false)
 	parse.ParseAssign();
 	parse.sc.MustGetNumber();
 
-	if ((parse.sc.Number >= 0 && parse.sc.Number <= 4) || parse.sc.Number == 8 || parse.sc.Number == 16)
+	if (parse.sc.Number == 8 || parse.sc.Number == 16) info->lightmode = ELightMode::NotSet;
+	else if (parse.sc.Number >= 0 && parse.sc.Number < 5)
 	{
 		info->lightmode = ELightMode(parse.sc.Number);
 	}
@@ -1459,6 +1590,57 @@ DEFINE_MAP_OPTION(lightmode, false)
 	}
 }
 
+DEFINE_MAP_OPTION(lightblendmode, false)
+{
+	parse.ParseAssign();
+	parse.sc.MustGetString();
+
+	if (parse.sc.Compare("Default") || parse.sc.Compare("Clamp"))
+	{
+		info->lightblendmode = ELightBlendMode::DEFAULT;
+	}
+	else if (parse.sc.Compare("ColoredClamp"))
+	{
+		info->lightblendmode = ELightBlendMode::CLAMP_COLOR;
+	}
+	else if (parse.sc.Compare("Unclamped"))
+	{
+		info->lightblendmode = ELightBlendMode::NOCLAMP;
+		if(parse.sc.CheckString(","))
+		{
+			parse.sc.MustGetString();
+			if (parse.sc.Compare("None"))
+			{
+				info->tonemap = ETonemapMode::None;
+			}
+			else if (parse.sc.Compare("Linear"))
+			{
+				info->tonemap = ETonemapMode::Linear;
+			}
+			else if (parse.sc.Compare("Uncharted2"))
+			{
+				info->tonemap = ETonemapMode::Uncharted2;
+			}
+			else if (parse.sc.Compare("HejlDawson"))
+			{
+				info->tonemap = ETonemapMode::HejlDawson;
+			}
+			else if (parse.sc.Compare("Reinhard"))
+			{
+				info->tonemap = ETonemapMode::Reinhard;
+			}
+			else
+			{
+				parse.sc.ScriptMessage("Invalid tonemap %s", parse.sc.String);
+			}
+		}
+	}
+	else
+	{
+		parse.sc.ScriptMessage("Invalid light blend mode %s", parse.sc.String);
+	}
+}
+
 DEFINE_MAP_OPTION(notexturefill, false)
 {
 	if (parse.CheckAssign())
@@ -1540,6 +1722,16 @@ DEFINE_MAP_OPTION(loadacs, false)
 	info->acsName = parse.sc.String;
 }
 
+DEFINE_MAP_OPTION(intro, true)
+{
+	parse.ParseCutscene(info->intro);
+}
+
+DEFINE_MAP_OPTION(outro, true)
+{
+	parse.ParseCutscene(info->outro);
+}
+
 
 //==========================================================================
 //
@@ -1561,6 +1753,7 @@ enum EMIType
 	MITYPE_CLRFLAG3,
 	MITYPE_SCFLAGS3,
 	MITYPE_COMPATFLAG,
+	MITYPE_CLRCOMPATFLAG,
 };
 
 struct MapInfoFlagHandler
@@ -1663,7 +1856,11 @@ MapFlagHandlers[] =
 	{ "enableskyboxao",					MITYPE_SETFLAG3,	LEVEL3_SKYBOXAO, 0 },
 	{ "disableskyboxao",				MITYPE_CLRFLAG3,	LEVEL3_SKYBOXAO, 0 },
 	{ "avoidmelee",						MITYPE_SETFLAG3,	LEVEL3_AVOIDMELEE, 0 },
+	{ "attenuatelights",				MITYPE_SETFLAG3,	LEVEL3_ATTENUATE, 0 },
+	{ "nofogofwar",					MITYPE_SETFLAG3,	LEVEL3_NOFOGOFWAR, 0 },
 	{ "nobotnodes",						MITYPE_IGNORE,	0, 0 },		// Skulltag option: nobotnodes
+	{ "nopassover",						MITYPE_COMPATFLAG, COMPATF_NO_PASSMOBJ, 0 },
+	{ "passover",						MITYPE_CLRCOMPATFLAG, COMPATF_NO_PASSMOBJ, 0 },
 	{ "compat_shorttex",				MITYPE_COMPATFLAG, COMPATF_SHORTTEX, 0 },
 	{ "compat_stairs",					MITYPE_COMPATFLAG, COMPATF_STAIRINDEX, 0 },
 	{ "compat_limitpain",				MITYPE_COMPATFLAG, COMPATF_LIMITPAIN, 0 },
@@ -1686,7 +1883,8 @@ MapFlagHandlers[] =
 	{ "compat_minotaur",				MITYPE_COMPATFLAG, COMPATF_MINOTAUR, 0 },
 	{ "compat_mushroom",				MITYPE_COMPATFLAG, COMPATF_MUSHROOM, 0 },
 	{ "compat_mbfmonstermove",			MITYPE_COMPATFLAG, COMPATF_MBFMONSTERMOVE, 0 },
-	{ "compat_corpsegibs",				MITYPE_COMPATFLAG, COMPATF_CORPSEGIBS, 0 },
+	{ "compat_corpsegibs",				MITYPE_COMPATFLAG, 0, 0 },	// this flag no longer exists, but we need it here for old mapinfos.
+	{ "compat_vileghosts",				MITYPE_COMPATFLAG, COMPATF_VILEGHOSTS, 0 },
 	{ "compat_noblockfriends",			MITYPE_COMPATFLAG, COMPATF_NOBLOCKFRIENDS, 0 },
 	{ "compat_spritesort",				MITYPE_COMPATFLAG, COMPATF_SPRITESORT, 0 },
 	{ "compat_light",					MITYPE_COMPATFLAG, COMPATF_LIGHT, 0 },
@@ -1706,6 +1904,9 @@ MapFlagHandlers[] =
 	{ "compat_scriptwait",				MITYPE_COMPATFLAG, 0, COMPATF2_SCRIPTWAIT },
 	{ "compat_avoidhazards",			MITYPE_COMPATFLAG, 0, COMPATF2_AVOID_HAZARDS },
 	{ "compat_stayonlift",				MITYPE_COMPATFLAG, 0, COMPATF2_STAYONLIFT },
+	{ "compat_nombf21",					MITYPE_COMPATFLAG, 0, COMPATF2_NOMBF21 },
+	{ "compat_voodoozombies",			MITYPE_COMPATFLAG, 0, COMPATF2_VOODOO_ZOMBIES },
+	{ "compat_noacsargcheck",			MITYPE_COMPATFLAG, 0, COMPATF2_NOACSARGCHECK },
 	{ "cd_start_track",					MITYPE_EATNEXT,	0, 0 },
 	{ "cd_end1_track",					MITYPE_EATNEXT,	0, 0 },
 	{ "cd_end2_track",					MITYPE_EATNEXT,	0, 0 },
@@ -1812,6 +2013,13 @@ void FMapInfoParser::ParseMapDefinition(level_info_t &info)
 				info.flags3 = (info.flags3 & handler->data2) | handler->data1;
 				break;
 
+			case MITYPE_CLRCOMPATFLAG:
+				info.compatflags &= ~handler->data1;
+				info.compatflags2 &= ~handler->data2;
+				info.compatmask |= handler->data1;
+				info.compatmask2 |= handler->data2;
+				break;
+
 			case MITYPE_COMPATFLAG:
 			{
 				int set = 1;
@@ -1996,7 +2204,7 @@ level_info_t *FMapInfoParser::ParseMapHeader(level_info_t &defaultinfo)
 			// This checks for a string labelled with the MapName and if that is identical to what got parsed here
 			// the string table entry will be used.
 
-			if (GStrings.MatchDefaultString(levelinfo->MapName, sc.String))
+			if (GStrings.MatchDefaultString(levelinfo->MapName.GetChars(), sc.String))
 			{
 				levelinfo->flags |= LEVEL_LOOKUPLEVELNAME;
 				levelinfo->LevelName = levelinfo->MapName;
@@ -2013,7 +2221,7 @@ level_info_t *FMapInfoParser::ParseMapHeader(level_info_t &defaultinfo)
 					if (fn && (!stricmp(fn, "HEXEN.WAD") || !stricmp(fn, "HEXDD.WAD")))
 					{
 						FStringf key("TXT_%.5s_%s", fn, levelinfo->MapName.GetChars());
-						if (GStrings.exists(key))
+						if (GStrings.exists(key.GetChars()))
 						{
 							levelinfo->flags |= LEVEL_LOOKUPLEVELNAME;
 							levelinfo->LevelName = key;
@@ -2026,7 +2234,7 @@ level_info_t *FMapInfoParser::ParseMapHeader(level_info_t &defaultinfo)
 
 	// Set up levelnum now so that you can use Teleport_NewMap specials
 	// to teleport to maps with standard names without needing a levelnum.
-	levelinfo->levelnum = GetDefaultLevelNum(levelinfo->MapName);
+	levelinfo->levelnum = GetDefaultLevelNum(levelinfo->MapName.GetChars());
 
 	// Does this map have a song defined via SNDINFO's $map command?
 	// Set that as this map's default music if it does.
@@ -2064,6 +2272,7 @@ void FMapInfoParser::ParseEpisodeInfo ()
 	bool noskill = false;
 	bool optional = false;
 	bool extended = false;
+	CutsceneDef introscene;
 
 	// Get map name
 	sc.MustGetString ();
@@ -2120,6 +2329,11 @@ void FMapInfoParser::ParseEpisodeInfo ()
 		{
 			noskill = true;
 		}
+		else if (sc.Compare("intro"))
+		{
+			ParseCutscene(introscene);
+		}
+
 		else if (!ParseCloseBrace())
 		{
 			// Unknown
@@ -2141,7 +2355,7 @@ void FMapInfoParser::ParseEpisodeInfo ()
 
 	if (optional && !remove)
 	{
-		if (!P_CheckMapData(map))
+		if (!P_CheckMapData(map.GetChars()))
 		{
 			// If the episode is optional and the map does not exist
 			// just ignore this episode definition.
@@ -2178,6 +2392,7 @@ void FMapInfoParser::ParseEpisodeInfo ()
 		epi->mPicName = pic;
 		epi->mShortcut = tolower(key);
 		epi->mNoSkill = noskill;
+		epi->mIntro = introscene;
 	}
 }
 
@@ -2242,9 +2457,11 @@ void FMapInfoParser::ParseMapInfo (int lump, level_info_t &gamedefaults, level_i
 						fileSystem.GetResourceFileFullName(fileSystem.GetFileContainer(inclump)), sc.String);
 				}
 			}
-			FScanner saved_sc = sc;
-			ParseMapInfo(inclump, gamedefaults, defaultinfo);
-			sc = saved_sc;
+			// use a new parser object to parse the include. Otherwise we'd have to save the entire FScanner in a local variable which is a lot more messy.
+			FMapInfoParser includer(&sc);
+			includer.format_type = format_type;
+			includer.HexenHack = HexenHack;
+			includer.ParseMapInfo(inclump, gamedefaults, defaultinfo);
 		}
 		else if (sc.Compare("gamedefaults"))
 		{
@@ -2422,13 +2639,64 @@ void G_ParseMapInfo (FString basemapinfo)
 {
 	int lump, lastlump = 0;
 	level_info_t gamedefaults;
+	TArray<FString> secretMaps;
+
+	int flags1 = 0, flags2 = 0;
+	if (gameinfo.gametype == GAME_Doom)
+	{
+		int comp = fileSystem.CheckNumForName("COMPLVL");
+		if (comp >= 0)
+		{
+			auto complvl = fileSystem.ReadFile(comp);
+			auto data = complvl.string();
+			int length = fileSystem.FileLength(comp);
+			if (length == 7 && !strnicmp("vanilla", data, 7))
+			{
+				flags1 = 
+					COMPATF_SHORTTEX | COMPATF_STAIRINDEX | COMPATF_USEBLOCKING | COMPATF_NODOORLIGHT | COMPATF_SPRITESORT |
+					COMPATF_TRACE | COMPATF_MISSILECLIP | COMPATF_SOUNDTARGET | COMPATF_DEHHEALTH | COMPATF_CROSSDROPOFF |
+					COMPATF_LIGHT | COMPATF_MASKEDMIDTEX |
+					COMPATF_LIMITPAIN | COMPATF_INVISIBILITY | COMPATF_VILEGHOSTS;
+
+				flags2 =
+					COMPATF2_FLOORMOVE | COMPATF2_EXPLODE1 | COMPATF2_NOMBF21 | COMPATF2_POINTONLINE;
+			}
+			else if (length == 4 && !strnicmp("boom", data, 4))
+			{
+				flags1 =
+					COMPATF_TRACE | COMPATF_SOUNDTARGET | COMPATF_BOOMSCROLL | COMPATF_MISSILECLIP | COMPATF_MASKEDMIDTEX |
+					COMPATF_INVISIBILITY;
+
+				flags2 =
+					COMPATF2_EXPLODE1 | COMPATF2_NOMBF21 | COMPATF2_POINTONLINE;
+			}
+			else if (length == 3 && !strnicmp("mbf", data, 3))
+			{
+				flags1 =
+					COMPATF_TRACE | COMPATF_SOUNDTARGET | COMPATF_BOOMSCROLL | COMPATF_MISSILECLIP | COMPATF_MUSHROOM |
+					COMPATF_MBFMONSTERMOVE | COMPATF_NOBLOCKFRIENDS | COMPATF_MASKEDMIDTEX | COMPATF_INVISIBILITY;
+
+				flags2 =
+					COMPATF2_EXPLODE1 | COMPATF2_AVOID_HAZARDS | COMPATF2_STAYONLIFT | COMPATF2_NOMBF21 | COMPATF2_POINTONLINE;
+			}
+			else if (length == 5 && !strnicmp("mbf21", data, 5))
+			{
+				flags1 =
+					COMPATF_TRACE | COMPATF_SOUNDTARGET | COMPATF_BOOMSCROLL | COMPATF_MISSILECLIP | COMPATF_MUSHROOM |
+					COMPATF_MASKEDMIDTEX | COMPATF_INVISIBILITY;
+
+				flags2 =
+					COMPATF2_EXPLODE1 | COMPATF2_AVOID_HAZARDS | COMPATF2_STAYONLIFT | COMPATF2_POINTONLINE;
+			}
+		}
+	}
 
 	// Parse the default MAPINFO for the current game. This lump *MUST* come from zdoom.pk3.
 	if (basemapinfo.IsNotEmpty())
 	{
 		FMapInfoParser parse;
 		level_info_t defaultinfo;
-		int baselump = fileSystem.GetNumForFullName(basemapinfo);
+		int baselump = fileSystem.GetNumForFullName(basemapinfo.GetChars());
 		if (fileSystem.GetFileContainer(baselump) > 0)
 		{
 			I_FatalError("File %s is overriding core lump %s.", 
@@ -2436,7 +2704,11 @@ void G_ParseMapInfo (FString basemapinfo)
 		}
 		parse.ParseMapInfo(baselump, gamedefaults, defaultinfo);
 	}
-
+	gamedefaults.compatflags |= flags1;
+	gamedefaults.compatmask |= flags1;
+	gamedefaults.compatflags2 |= flags2;
+	gamedefaults.compatmask2 |= flags2;
+	
 	static const char *mapinfonames[] = { "MAPINFO", "ZMAPINFO", "UMAPINFO", NULL };
 	int nindex;
 
@@ -2449,7 +2721,7 @@ void G_ParseMapInfo (FString basemapinfo)
 			// If that exists we need to skip this one.
 
 			int wad = fileSystem.GetFileContainer(lump);
-			int altlump = fileSystem.CheckNumForName("ZMAPINFO", ns_global, wad, true);
+			int altlump = fileSystem.CheckNumForName("ZMAPINFO", FileSys::ns_global, wad, true);
 
 			if (altlump >= 0) continue;
 		}
@@ -2457,9 +2729,9 @@ void G_ParseMapInfo (FString basemapinfo)
 		{
 			// MAPINFO and ZMAPINFO will override UMAPINFO if in the same WAD.
 			int wad = fileSystem.GetFileContainer(lump);
-			int altlump = fileSystem.CheckNumForName("ZMAPINFO", ns_global, wad, true);
+			int altlump = fileSystem.CheckNumForName("ZMAPINFO", FileSys::ns_global, wad, true);
 			if (altlump >= 0) continue;
-			altlump = fileSystem.CheckNumForName("MAPINFO", ns_global, wad, true);
+			altlump = fileSystem.CheckNumForName("MAPINFO", FileSys::ns_global, wad, true);
 			if (altlump >= 0) continue;
 		}
 		if (nindex != 2)
@@ -2484,5 +2756,23 @@ void G_ParseMapInfo (FString basemapinfo)
 	{
 		I_FatalError ("You cannot use clearskills in a MAPINFO if you do not define any new skills after it.");
 	}
+
+	// Find any and all secret maps.
+	for (unsigned int i = 0; i < wadlevelinfos.Size(); i++)
+	{
+		if (wadlevelinfos[i].NextSecretMap.IsNotEmpty() && wadlevelinfos[i].NextSecretMap != wadlevelinfos[i].NextMap)
+		{
+			secretMaps.Push(wadlevelinfos[i].NextSecretMap);
+		}
+	}
+	// ...and then mark them all as secret maps.
+	for (unsigned int i = 0; i < secretMaps.Size(); i++)
+	{
+		auto* li = FindLevelInfo(secretMaps[i].GetChars(), false);
+		if (li)
+		{
+			li->flags3 |= LEVEL3_SECRET;
+		}
+	}
 }
 
