@@ -56,7 +56,7 @@
 #include "g_level.h"
 #include "d_event.h"
 #include "p_tick.h"
-#include "st_start.h"
+#include "startscreen.h"
 #include "d_main.h"
 #include "i_system.h"
 #include "doommenu.h"
@@ -64,6 +64,9 @@
 #include "gameconfigfile.h"
 #include "d_player.h"
 #include "teaminfo.h"
+#include "i_time.h"
+#include "shiftstate.h"
+#include "s_music.h"
 #include "hwrenderer/scene/hw_drawinfo.h"
 
 EXTERN_CVAR(Int, cl_gfxlocalization)
@@ -71,17 +74,78 @@ EXTERN_CVAR(Bool, m_quickexit)
 EXTERN_CVAR(Bool, saveloadconfirmation) // [mxd]
 EXTERN_CVAR(Bool, quicksaverotation)
 EXTERN_CVAR(Bool, show_messages)
+EXTERN_CVAR(Float, hud_scalefactor)
 
 CVAR(Bool, m_simpleoptions, true, CVAR_ARCHIVE|CVAR_GLOBALCONFIG)
 
 typedef void(*hfunc)();
 DMenu* CreateMessageBoxMenu(DMenu* parent, const char* message, int messagemode, bool playsound, FName action = NAME_None, hfunc handler = nullptr);
 bool OkForLocalization(FTextureID texnum, const char* substitute);
-void I_WaitVBL(int count);
 
 
 FNewGameStartup NewGameStartupInfo;
+int LastSkill = -1;
 
+void StartGameDirect(bool hasPlayerClass, bool randomPlayerClass, PClassActor * playerClass, int Episode, int Skill)
+{
+	// shouldn't work outside of a menu
+	if (DMenu::InMenu)
+	{
+		if(!netgame)
+		{
+			NewGameStartupInfo.hasPlayerClass = hasPlayerClass;
+
+			if(hasPlayerClass)
+			{
+				if(randomPlayerClass)
+				{
+					NewGameStartupInfo.PlayerClass = "Random";
+				}
+				else if(!playerClass)
+				{
+					NullParam("playerClass");
+				}
+				else
+				{
+					NewGameStartupInfo.PlayerClass = playerClass->GetDisplayName();
+				}
+			}
+
+			NewGameStartupInfo.Episode = Episode;
+			NewGameStartupInfo.Skill = Skill;
+
+			G_DeferedInitNew (&NewGameStartupInfo);
+
+			if (gamestate == GS_FULLCONSOLE)
+			{
+				gamestate = GS_HIDECONSOLE;
+				gameaction = ga_newgame;
+			}
+		}
+		else
+		{
+			DPrintf(DMSG_WARNING, TEXTCOLOR_RED "Cannot start a new game during a netgame\n");
+		}
+
+		M_ClearMenus ();
+	}
+	else
+	{
+		ThrowAbortException(X_OTHER, "Attempt to start a new game outside of menu code");
+	}
+}
+
+DEFINE_ACTION_FUNCTION_NATIVE(DMenu, StartGameDirect, StartGameDirect)
+{
+	PARAM_PROLOGUE;
+	PARAM_BOOL(hasPlayerClass);
+	PARAM_BOOL(randomPlayerClass);
+	PARAM_POINTER(playerClass, PClassActor);
+	PARAM_INT(Episode);
+	PARAM_INT(Skill);
+	StartGameDirect(hasPlayerClass, randomPlayerClass, playerClass, Episode, Skill);
+	return 0;
+}
 
 bool M_SetSpecialMenu(FName& menu, int param)
 {
@@ -123,9 +187,17 @@ bool M_SetSpecialMenu(FName& menu, int param)
 		// sent from the player class menu
 		NewGameStartupInfo.Skill = -1;
 		NewGameStartupInfo.Episode = -1;
-		NewGameStartupInfo.PlayerClass = 
-			param == -1000? nullptr :
-			param == -1? "Random" : GetPrintableDisplayName(PlayerClasses[param].Type).GetChars();
+
+		if(param == -1000)
+		{
+			NewGameStartupInfo.hasPlayerClass = false;
+		}
+		else
+		{
+			NewGameStartupInfo.hasPlayerClass = true;
+			NewGameStartupInfo.PlayerClass = (param == -1) ? FString("Random") : PlayerClasses[param].Type->GetDisplayName();
+		}
+
 		M_StartupEpisodeMenu(&NewGameStartupInfo);	// needs player class name from class menu (later)
 		break;
 
@@ -135,7 +207,7 @@ bool M_SetSpecialMenu(FName& menu, int param)
 		if ((gameinfo.flags & GI_SHAREWARE) && param > 0)
 		{
 			// Only Doom and Heretic have multi-episode shareware versions.
-			M_StartMessage(GStrings("SWSTRING"), 1);
+			M_StartMessage(GStrings.GetString("SWSTRING"), 1);
 			return false;
 		}
 
@@ -147,9 +219,10 @@ bool M_SetSpecialMenu(FName& menu, int param)
 	{
 		// sent from the skill menu for a skill that needs to be confirmed
 		NewGameStartupInfo.Skill = param;
+		LastSkill = param;
 
-		const char *msg = AllSkills[param].MustConfirmText;
-		if (*msg==0) msg = GStrings("NIGHTMARE");
+		const char *msg = AllSkills[param].MustConfirmText.GetChars();
+		if (*msg==0) msg = GStrings.GetString("NIGHTMARE");
 		M_StartMessage (msg, 0, NAME_StartgameConfirmed);
 		return false;
 	}
@@ -157,6 +230,7 @@ bool M_SetSpecialMenu(FName& menu, int param)
 	case NAME_Startgame:
 		// sent either from skill menu or confirmation screen. Skill gets only set if sent from skill menu
 		// Now we can finally start the game. Ugh...
+		LastSkill = param;
 		NewGameStartupInfo.Skill = param;
 		[[fallthrough]];
 	case NAME_StartgameConfirmed:
@@ -174,7 +248,7 @@ bool M_SetSpecialMenu(FName& menu, int param)
 		if (!usergame || (players[consoleplayer].health <= 0 && !multiplayer) || gamestate != GS_LEVEL)
 		{
 			// cannot save outside the game.
-			M_StartMessage (GStrings("SAVEDEAD"), 1);
+			M_StartMessage (GStrings.GetString("SAVEDEAD"), 1);
 			return false;
 		}
 		break;
@@ -202,6 +276,10 @@ bool M_SetSpecialMenu(FName& menu, int param)
 		menu = NAME_Optionsmenu;
 		break;
 
+	case NAME_Readthismenu:
+		// [MK] allow us to override the ReadThisMenu class
+		menu = gameinfo.HelpMenuClass;
+		break;
 	}
 
 	DMenuDescriptor** desc = MenuDescriptors.CheckKey(menu);
@@ -209,7 +287,7 @@ bool M_SetSpecialMenu(FName& menu, int param)
 	{
 		if ((*desc)->mNetgameMessage.IsNotEmpty() && netgame && !demoplayback)
 		{
-			M_StartMessage((*desc)->mNetgameMessage, 1);
+			M_StartMessage((*desc)->mNetgameMessage.GetChars(), 1);
 			return false;
 		}
 	}
@@ -225,7 +303,7 @@ bool M_SetSpecialMenu(FName& menu, int param)
 //
 //=============================================================================
 
-void M_StartControlPanel(bool makeSound, bool scaleoverride)
+void OnMenuOpen(bool makeSound)
 {
 	if (hud_toggled)
 		D_ToggleHud();
@@ -240,7 +318,6 @@ void M_StartControlPanel(bool makeSound, bool scaleoverride)
 	{
 		S_Sound(CHAN_VOICE, CHANF_UI, "menu/activate", snd_menuvolume, ATTN_NONE);
 	}
-	M_DoStartControlPanel(scaleoverride);
 }
 
 
@@ -286,6 +363,15 @@ void System_M_Dim()
 }
 
 
+static void M_Quit()
+{
+	DeleteScreenJob();
+	S_StopAllChannels();
+	S_StopMusic(true);
+	CleanSWDrawer();
+	ST_Endoom();
+}
+
 //=============================================================================
 //
 //
@@ -296,29 +382,28 @@ CCMD (menu_quit)
 {	// F10
 	if (m_quickexit)
 	{
-		CleanSWDrawer();
-		ST_Endoom();
+		M_Quit();
 	}
 
 	M_StartControlPanel (true);
 
 	const size_t messageindex = static_cast<size_t>(gametic) % gameinfo.quitmessages.Size();
 	FString EndString;
-	const char *msg = gameinfo.quitmessages[messageindex];
+	const char *msg = gameinfo.quitmessages[messageindex].GetChars();
 	if (msg[0] == '$')
 	{
 		if (msg[1] == '*')
 		{
-			EndString = GStrings(msg + 2);
+			EndString = GStrings.GetString(msg + 2);
 		}
 		else
 		{
-			EndString.Format("%s\n\n%s", GStrings(msg + 1), GStrings("DOSY"));
+			EndString.Format("%s\n\n%s", GStrings.GetString(msg + 1), GStrings.GetString("DOSY"));
 		}
 	}
 	else EndString = gameinfo.quitmessages[messageindex];
 
-	DMenu *newmenu = CreateMessageBoxMenu(CurrentMenu, EndString, 0, false, NAME_None, []()
+	DMenu *newmenu = CreateMessageBoxMenu(CurrentMenu, EndString.GetChars(), 0, false, NAME_None, []()
 	{
 		if (!netgame)
 		{
@@ -328,8 +413,7 @@ CCMD (menu_quit)
 				I_WaitVBL(105);
 			}
 		}
-		CleanSWDrawer();
-		ST_Endoom();
+		M_Quit();
 	});
 
 
@@ -346,8 +430,8 @@ CCMD (menu_quit)
 
 void ActivateEndGameMenu()
 {
-	FString tempstring = GStrings(netgame ? "NETEND" : "ENDGAME");
-	DMenu *newmenu = CreateMessageBoxMenu(CurrentMenu, tempstring, 0, false, NAME_None, []()
+	FString tempstring = GStrings.GetString(netgame ? "NETEND" : "ENDGAME");
+	DMenu *newmenu = CreateMessageBoxMenu(CurrentMenu, tempstring.GetChars(), 0, false, NAME_None, []()
 	{
 		M_ClearMenus();
 		if (!netgame)
@@ -416,10 +500,10 @@ CCMD (quicksave)
 
 	S_Sound(CHAN_VOICE, CHANF_UI, "menu/activate", snd_menuvolume, ATTN_NONE);
 
-	FString tempstring = GStrings("QSPROMPT");
+	FString tempstring = GStrings.GetString("QSPROMPT");
 	tempstring.Substitute("%s", savegameManager.quickSaveSlot->SaveTitle.GetChars());
 
-	DMenu *newmenu = CreateMessageBoxMenu(CurrentMenu, tempstring, 0, false, NAME_None, []()
+	DMenu *newmenu = CreateMessageBoxMenu(CurrentMenu, tempstring.GetChars(), 0, false, NAME_None, []()
 	{
 		G_SaveGame(savegameManager.quickSaveSlot->Filename.GetChars(), savegameManager.quickSaveSlot->SaveTitle.GetChars());
 		S_Sound(CHAN_VOICE, CHANF_UI, "menu/dismiss", snd_menuvolume, ATTN_NONE);
@@ -440,7 +524,7 @@ CCMD (quickload)
 	if (netgame)
 	{
 		M_StartControlPanel(true);
-		M_StartMessage (GStrings("QLOADNET"), 1);
+		M_StartMessage (GStrings.GetString("QLOADNET"), 1);
 		return;
 	}
 		
@@ -459,12 +543,12 @@ CCMD (quickload)
 		G_LoadGame(savegameManager.quickSaveSlot->Filename.GetChars());
 		return;
 	}
-	FString tempstring = GStrings("QLPROMPT");
+	FString tempstring = GStrings.GetString("QLPROMPT");
 	tempstring.Substitute("%s", savegameManager.quickSaveSlot->SaveTitle.GetChars());
 
 	M_StartControlPanel(true);
 
-	DMenu *newmenu = CreateMessageBoxMenu(CurrentMenu, tempstring, 0, false, NAME_None, []()
+	DMenu *newmenu = CreateMessageBoxMenu(CurrentMenu, tempstring.GetChars(), 0, false, NAME_None, []()
 	{
 		G_LoadGame(savegameManager.quickSaveSlot->Filename.GetChars());
 		S_Sound(CHAN_VOICE, CHANF_UI, "menu/dismiss", snd_menuvolume, ATTN_NONE);
@@ -483,13 +567,13 @@ CCMD (togglemessages)
 {
 	if (show_messages)
 	{
-		Printf(TEXTCOLOR_RED "%s\n", GStrings("MSGOFF"));
+		Printf(TEXTCOLOR_RED "%s\n", GStrings.GetString("MSGOFF"));
 		show_messages = false;
 	}
 	else
 	{
 		show_messages = true;
-		Printf(TEXTCOLOR_RED "%s\n", GStrings("MSGON"));
+		Printf(TEXTCOLOR_RED "%s\n", GStrings.GetString("MSGON"));
 	}
 }
 
@@ -497,14 +581,28 @@ EXTERN_CVAR (Int, screenblocks)
 
 CCMD (sizedown)
 {
-	screenblocks = screenblocks - 1;
+	if (shiftState.ShiftPressed())
+	{
+		hud_scalefactor = hud_scalefactor - 0.04f;
+	}
+	else
+	{
+		screenblocks = screenblocks - 1;
+	}
 	S_Sound (CHAN_VOICE, CHANF_UI, "menu/change", snd_menuvolume, ATTN_NONE);
 }
 
 CCMD (sizeup)
 {
-	screenblocks = screenblocks + 1;
-	S_Sound (CHAN_VOICE, CHANF_UI, "menu/change", snd_menuvolume, ATTN_NONE);
+	if (shiftState.ShiftPressed())
+	{
+		hud_scalefactor = hud_scalefactor + 0.04f;
+	}
+	else
+	{
+		screenblocks = screenblocks + 1;
+	}
+	S_Sound(CHAN_VOICE, CHANF_UI, "menu/change", snd_menuvolume, ATTN_NONE);
 }
 
 CCMD(reset2defaults)
@@ -517,8 +615,8 @@ CCMD(reset2defaults)
 CCMD(reset2saved)
 {
 	GameConfig->DoGlobalSetup ();
-	GameConfig->DoGameSetup (gameinfo.ConfigName);
-	GameConfig->DoModSetup (gameinfo.ConfigName);
+	GameConfig->DoGameSetup (gameinfo.ConfigName.GetChars());
+	GameConfig->DoModSetup (gameinfo.ConfigName.GetChars());
 	R_SetViewSize (screenblocks);
 }
 
@@ -576,8 +674,8 @@ void M_StartupEpisodeMenu(FNewGameStartup *gs)
 			{
 				if (AllEpisodes[i].mPicName.IsNotEmpty())
 				{
-					FTextureID tex = GetMenuTexture(AllEpisodes[i].mPicName);
-					if (AllEpisodes[i].mEpisodeName.IsEmpty() || OkForLocalization(tex, AllEpisodes[i].mEpisodeName))
+					FTextureID tex = GetMenuTexture(AllEpisodes[i].mPicName.GetChars());
+					if (AllEpisodes[i].mEpisodeName.IsEmpty() || OkForLocalization(tex, AllEpisodes[i].mEpisodeName.GetChars()))
 						continue;
 				}
 				if ((gameinfo.gametype & GAME_DoomStrifeChex) && spacing == 16) spacing = 18;
@@ -587,9 +685,9 @@ void M_StartupEpisodeMenu(FNewGameStartup *gs)
 			// center the menu on the screen if the top space is larger than the bottom space
 			int totalheight = posy + AllEpisodes.Size() * spacing - topy;
 
-			if (totalheight < 190 || AllEpisodes.Size() == 1)
+			if (ld->mForceList || totalheight < 190 || AllEpisodes.Size() == 1)
 			{
-				int newtop = (200 - totalheight) / 2;
+				int newtop = max(10, 200 - totalheight) / 2;
 				int topdelta = newtop - topy;
 				if (topdelta < 0)
 				{
@@ -608,15 +706,15 @@ void M_StartupEpisodeMenu(FNewGameStartup *gs)
 					DMenuItemBase *it = nullptr;
 					if (AllEpisodes[i].mPicName.IsNotEmpty())
 					{
-						FTextureID tex = GetMenuTexture(AllEpisodes[i].mPicName);
-						if (AllEpisodes[i].mEpisodeName.IsEmpty() || OkForLocalization(tex, AllEpisodes[i].mEpisodeName))
+						FTextureID tex = GetMenuTexture(AllEpisodes[i].mPicName.GetChars());
+						if (AllEpisodes[i].mEpisodeName.IsEmpty() || OkForLocalization(tex, AllEpisodes[i].mEpisodeName.GetChars()))
 							continue;	// We do not measure patch based entries. They are assumed to fit
 					}
-					const char *c = AllEpisodes[i].mEpisodeName;
-					if (*c == '$') c = GStrings(c + 1);
+					const char *c = AllEpisodes[i].mEpisodeName.GetChars();
+					if (*c == '$') c = GStrings.GetString(c + 1);
 					int textwidth = ld->mFont->StringWidth(c);
 					int textright = posx + textwidth;
-					if (posx + textright > 320) posx = std::max(0, 320 - textright);
+					if (posx + textright > 320) posx = max(0, 320 - textright);
 				}
 
 				for(unsigned i = 0; i < AllEpisodes.Size(); i++)
@@ -624,14 +722,14 @@ void M_StartupEpisodeMenu(FNewGameStartup *gs)
 					DMenuItemBase *it = nullptr;
 					if (AllEpisodes[i].mPicName.IsNotEmpty())
 					{
-						FTextureID tex = GetMenuTexture(AllEpisodes[i].mPicName);
-						if (AllEpisodes[i].mEpisodeName.IsEmpty() || OkForLocalization(tex, AllEpisodes[i].mEpisodeName))
+						FTextureID tex = GetMenuTexture(AllEpisodes[i].mPicName.GetChars());
+						if (AllEpisodes[i].mEpisodeName.IsEmpty() || OkForLocalization(tex, AllEpisodes[i].mEpisodeName.GetChars()))
 							it = CreateListMenuItemPatch(posx, posy, spacing, AllEpisodes[i].mShortcut, tex, NAME_Skillmenu, i);
 					}
 					if (it == nullptr)
 					{
 						it = CreateListMenuItemText(posx, posy, spacing, AllEpisodes[i].mShortcut, 
-							AllEpisodes[i].mEpisodeName, ld->mFont, ld->mFontColor, ld->mFontColor2, NAME_Skillmenu, i);
+							AllEpisodes[i].mEpisodeName.GetChars(), ld->mFont, ld->mFontColor, ld->mFontColor2, NAME_Skillmenu, i);
 					}
 					ld->mItems.Push(it);
 					posy += spacing;
@@ -665,10 +763,13 @@ void M_StartupEpisodeMenu(FNewGameStartup *gs)
 		od->mScrollTop = 0;
 		od->mIndent = 160;
 		od->mDontDim = false;
+		od->mDontBlur = false;
+		od->mAnimatedTransition = false;
+		od->mAnimated = false;
 		GC::WriteBarrier(od);
 		for(unsigned i = 0; i < AllEpisodes.Size(); i++)
 		{
-			auto it = CreateOptionMenuItemSubmenu(AllEpisodes[i].mEpisodeName, "Skillmenu", i);
+			auto it = CreateOptionMenuItemSubmenu(AllEpisodes[i].mEpisodeName.GetChars(), "Skillmenu", i);
 			od->mItems.Push(it);
 			GC::WriteBarrier(od, it);
 		}
@@ -712,7 +813,7 @@ static void BuildPlayerclassMenu()
 			{
 				if (!(PlayerClasses[i].Flags & PCF_NOMENU))
 				{
-					const char *pname = GetPrintableDisplayName(PlayerClasses[i].Type);
+					const char *pname = GetPrintableDisplayName(PlayerClasses[i].Type).GetChars();
 					if (pname != nullptr)
 					{
 						numclassitems++;
@@ -731,9 +832,9 @@ static void BuildPlayerclassMenu()
 				ld->mAutoselect = ld->mItems.Push(it);
 				success = true;
 			}
-			else if (totalheight <= 190)
+			else if (ld->mForceList || totalheight <= 190)
 			{
-				int newtop = (200 - totalheight + topy) / 2;
+				int newtop = (max(10, 200 - totalheight) + topy) / 2;
 				int topdelta = newtop - topy;
 				if (topdelta < 0)
 				{
@@ -749,7 +850,7 @@ static void BuildPlayerclassMenu()
 				{
 					if (!(PlayerClasses[i].Flags & PCF_NOMENU))
 					{
-						const char *pname = GetPrintableDisplayName(PlayerClasses[i].Type);
+						const char *pname = GetPrintableDisplayName(PlayerClasses[i].Type).GetChars();
 						if (pname != nullptr)
 						{
 							auto it = CreateListMenuItemText(ld->mXpos, ld->mYpos, ld->mLinespacing, *pname,
@@ -768,7 +869,7 @@ static void BuildPlayerclassMenu()
 				}
 				if (n == 0)
 				{
-					const char *pname = GetPrintableDisplayName(PlayerClasses[0].Type);
+					const char *pname = GetPrintableDisplayName(PlayerClasses[0].Type).GetChars();
 					if (pname != nullptr)
 					{
 						auto it = CreateListMenuItemText(ld->mXpos, ld->mYpos, ld->mLinespacing, *pname,
@@ -800,13 +901,16 @@ static void BuildPlayerclassMenu()
 		od->mScrollTop = 0;
 		od->mIndent = 160;
 		od->mDontDim = false;
+		od->mDontBlur = false;
+		od->mAnimatedTransition = false;
+		od->mAnimated = false;
 		od->mNetgameMessage = "$NEWGAME";
 		GC::WriteBarrier(od);
 		for (unsigned i = 0; i < PlayerClasses.Size (); i++)
 		{
 			if (!(PlayerClasses[i].Flags & PCF_NOMENU))
 			{
-				const char *pname = GetPrintableDisplayName(PlayerClasses[i].Type);
+				const char *pname = GetPrintableDisplayName(PlayerClasses[i].Type).GetChars();
 				if (pname != nullptr)
 				{
 					auto it = CreateOptionMenuItemSubmenu(pname, "Episodemenu", i);
@@ -898,12 +1002,12 @@ static void InitKeySections()
 				FKeySection *sect = &KeySections[i];
 				DMenuItemBase *item = CreateOptionMenuItemStaticText(" ");
 				menu->mItems.Push(item);
-				item = CreateOptionMenuItemStaticText(sect->mTitle, 1);
+				item = CreateOptionMenuItemStaticText(sect->mTitle.GetChars(), 1);
 				menu->mItems.Push(item);
 				for (unsigned j = 0; j < sect->mActions.Size(); j++)
 				{
 					FKeyAction *act = &sect->mActions[j];
-					item = CreateOptionMenuItemControl(act->mTitle, act->mAction, &Bindings);
+					item = CreateOptionMenuItemControl(act->mTitle.GetChars(), act->mAction, &Bindings);
 					menu->mItems.Push(item);
 				}
 			}
@@ -1044,7 +1148,7 @@ void M_StartupSkillMenu(FNewGameStartup *gs)
 	}
 	if (MenuSkills.Size() == 0) I_Error("No valid skills for menu found. At least one must be defined.");
 
-	int defskill = DefaultSkill;
+	int defskill = LastSkill > -1? LastSkill : DefaultSkill; // use the last selected skill, if available.
 	if ((unsigned int)defskill >= MenuSkills.Size())
 	{
 		defskill = SkillIndices[(MenuSkills.Size() - 1) / 2];
@@ -1106,8 +1210,8 @@ void M_StartupSkillMenu(FNewGameStartup *gs)
 				{
 					if (MenuSkills[i]->PicName.IsNotEmpty())
 					{
-						FTextureID tex = GetMenuTexture(MenuSkills[i]->PicName);
-						if (MenuSkills[i]->MenuName.IsEmpty() || OkForLocalization(tex, MenuSkills[i]->MenuName))
+						FTextureID tex = GetMenuTexture(MenuSkills[i]->PicName.GetChars());
+						if (MenuSkills[i]->MenuName.IsEmpty() || OkForLocalization(tex, MenuSkills[i]->MenuName.GetChars()))
 							continue;
 					}
 					if ((gameinfo.gametype & GAME_DoomStrifeChex) && spacing == 16) spacing = 18;
@@ -1117,9 +1221,9 @@ void M_StartupSkillMenu(FNewGameStartup *gs)
 				// center the menu on the screen if the top space is larger than the bottom space
 				int totalheight = posy + MenuSkills.Size() * spacing - topy;
 
-				if (totalheight < 190 || MenuSkills.Size() == 1)
+				if (ld->mForceList || totalheight < 190 || MenuSkills.Size() == 1)
 				{
-					int newtop = (200 - totalheight) / 2;
+					int newtop = max(10, 200 - totalheight) / 2;
 					int topdelta = newtop - topy;
 					if (topdelta < 0)
 					{
@@ -1145,22 +1249,22 @@ void M_StartupSkillMenu(FNewGameStartup *gs)
 				DMenuItemBase *li = nullptr;
 
 				FString *pItemText = nullptr;
-				if (gs->PlayerClass != nullptr)
+				if (gs->hasPlayerClass)
 				{
 					pItemText = skill.MenuNamesForPlayerClass.CheckKey(gs->PlayerClass);
 				}
 
 				if (skill.PicName.Len() != 0 && pItemText == nullptr)
 				{
-					FTextureID tex = GetMenuTexture(skill.PicName);
-					if (skill.MenuName.IsEmpty() || OkForLocalization(tex, skill.MenuName))
+					FTextureID tex = GetMenuTexture(skill.PicName.GetChars());
+					if (skill.MenuName.IsEmpty() || OkForLocalization(tex, skill.MenuName.GetChars()))
 						continue;
 				}
 				const char *c = pItemText ? pItemText->GetChars() : skill.MenuName.GetChars();
-				if (*c == '$') c = GStrings(c + 1);
+				if (*c == '$') c = GStrings.GetString(c + 1);
 				int textwidth = ld->mFont->StringWidth(c);
 				int textright = posx + textwidth;
-				if (posx + textright > 320) posx = std::max(0, 320 - textright);
+				if (posx + textright > 320) posx = max(0, 320 - textright);
 			}
 
 			unsigned firstitem = ld->mItems.Size();
@@ -1172,7 +1276,7 @@ void M_StartupSkillMenu(FNewGameStartup *gs)
 				FName action = (skill.MustConfirm && !AllEpisodes[gs->Episode].mNoSkill) ?
 					NAME_StartgameConfirm : NAME_Startgame;
 				FString *pItemText = nullptr;
-				if (gs->PlayerClass != nullptr)
+				if (gs->hasPlayerClass)
 				{
 					pItemText = skill.MenuNamesForPlayerClass.CheckKey(gs->PlayerClass);
 				}
@@ -1181,14 +1285,14 @@ void M_StartupSkillMenu(FNewGameStartup *gs)
 				if (color == CR_UNTRANSLATED) color = ld->mFontColor;
 				if (skill.PicName.Len() != 0 && pItemText == nullptr)
 				{
-					FTextureID tex = GetMenuTexture(skill.PicName);
-					if (skill.MenuName.IsEmpty() || OkForLocalization(tex, skill.MenuName))
+					FTextureID tex = GetMenuTexture(skill.PicName.GetChars());
+					if (skill.MenuName.IsEmpty() || OkForLocalization(tex, skill.MenuName.GetChars()))
 						li = CreateListMenuItemPatch(posx, y, spacing, skill.Shortcut, tex, action, SkillIndices[i]);
 				}
 				if (li == nullptr)
 				{
 					li = CreateListMenuItemText(posx, y, spacing, skill.Shortcut,
-									pItemText? *pItemText : skill.MenuName, ld->mFont, color,ld->mFontColor2, action, SkillIndices[i]);
+									pItemText? pItemText->GetChars() : skill.MenuName.GetChars(), ld->mFont, color, ld->mFontColor2, action, SkillIndices[i]);
 				}
 				ld->mItems.Push(li);
 				GC::WriteBarrier(*desc, li);
@@ -1223,6 +1327,9 @@ fail:
 		od->mScrollTop = 0;
 		od->mIndent = 160;
 		od->mDontDim = false;
+		od->mDontBlur = false;
+		od->mAnimatedTransition = false;
+		od->mAnimated = false;
 		GC::WriteBarrier(od);
 	}
 	else
@@ -1239,11 +1346,11 @@ fail:
 			"StartgameConfirm" : "Startgame";
 
 		FString *pItemText = nullptr;
-		if (gs->PlayerClass != nullptr)
+		if (gs->hasPlayerClass)
 		{
 			pItemText = skill.MenuNamesForPlayerClass.CheckKey(gs->PlayerClass);
 		}
-		li = CreateOptionMenuItemSubmenu(pItemText? *pItemText : skill.MenuName, action, SkillIndices[i]);
+		li = CreateOptionMenuItemSubmenu(pItemText? pItemText->GetChars() : skill.MenuName.GetChars(), action, SkillIndices[i]);
 		od->mItems.Push(li);
 		GC::WriteBarrier(od, li);
 		if (!done)
@@ -1300,7 +1407,17 @@ void SetDefaultMenuColors()
 	OptionSettings.mFontColorHighlight = V_FindFontColor(gameinfo.mFontColorHighlight);
 	OptionSettings.mFontColorSelection = V_FindFontColor(gameinfo.mFontColorSelection);
 
-	auto cls = PClass::FindClass("DoomMenuDelegate");
+	auto cls = PClass::FindClass(gameinfo.HelpMenuClass);
+	if (!cls)
+		I_FatalError("%s: Undefined help menu class", gameinfo.HelpMenuClass.GetChars());
+	if (!cls->IsDescendantOf(RUNTIME_CLASS(DMenu)))
+		I_FatalError("'%s' does not inherit from Menu", gameinfo.HelpMenuClass.GetChars());
+
+	cls = PClass::FindClass(gameinfo.MenuDelegateClass);
+	if (!cls)
+		I_FatalError("%s: Undefined menu delegate class", gameinfo.MenuDelegateClass.GetChars());
+	if (!cls->IsDescendantOf("MenuDelegateBase"))
+		I_FatalError("'%s' does not inherit from MenuDelegateBase", gameinfo.MenuDelegateClass.GetChars());
 	menuDelegate = cls->CreateNew();
 }
 
@@ -1430,31 +1547,3 @@ CCMD (menu_video)
 	M_StartControlPanel (true);
 	M_SetMenu(NAME_VideoModeMenu, -1);
 }
-
-
-#ifdef _WIN32
-EXTERN_CVAR(Bool, vr_enable_quadbuffered)
-#endif
-
-void UpdateVRModes(bool considerQuadBuffered)
-{
-	FOptionValues** pVRModes = OptionValues.CheckKey("VRMode");
-	if (pVRModes == nullptr) return;
-
-	TArray<FOptionValues::Pair>& vals = (*pVRModes)->mValues;
-	TArray<FOptionValues::Pair> filteredValues;
-	int cnt = vals.Size();
-	for (int i = 0; i < cnt; ++i) {
-		auto const& mode = vals[i];
-		if (mode.Value == 7) {  // Quad-buffered stereo
-#ifdef _WIN32
-			if (!vr_enable_quadbuffered) continue;
-#else
-			continue;  // Remove quad-buffered option on Mac and Linux
-#endif
-			if (!considerQuadBuffered) continue;  // Probably no compatible screen mode was found
-		}
-		filteredValues.Push(mode);
-	}
-	vals = filteredValues;
-}
