@@ -43,6 +43,10 @@
 #include "i_time.h"
 #include "texturemanager.h"
 #include "modelrenderer.h"
+#include "actor.h"
+#include "actorinlines.h"
+#include "v_video.h"
+#include "hw_bonebuffer.h"
 
 
 #ifdef _MSC_VER
@@ -55,15 +59,17 @@ EXTERN_CVAR (Bool, r_drawvoxels)
 extern TDeletingArray<FVoxel *> Voxels;
 extern TDeletingArray<FVoxelDef *> VoxelDefs;
 
-void RenderFrameModels(FModelRenderer* renderer, FLevelLocals* Level, const FSpriteModelFrame* smf, const FState* curState, const int curTics, const PClass* ti, int translation);
+void RenderFrameModels(FModelRenderer* renderer, FLevelLocals* Level, const FSpriteModelFrame *smf, const FState* curState, const int curTics, FTranslationID translation, AActor* actor);
 
 
 void RenderModel(FModelRenderer *renderer, float x, float y, float z, FSpriteModelFrame *smf, AActor *actor, double ticFrac)
 {
 	// Setup transformation.
 
-	int translation = 0;
-	if (!(smf->flags & MDL_IGNORETRANSLATION))
+	int smf_flags = smf->getFlags(actor->modelData);
+
+	FTranslationID translation = NO_TRANSLATION;
+	if (!(smf_flags & MDL_IGNORETRANSLATION))
 		translation = actor->Translation;
 
 	// y scale for a sprite means height, i.e. z in the world!
@@ -78,10 +84,10 @@ void RenderModel(FModelRenderer *renderer, float x, float y, float z, FSpriteMod
 		angles = actor->InterpolatedAngles(ticFrac);
 	else
 		angles = actor->Angles;
-	float angle = angles.Yaw.Degrees;
+	float angle = angles.Yaw.Degrees();
 
 	// [BB] Workaround for the missing pitch information.
-	if ((smf->flags & MDL_PITCHFROMMOMENTUM))
+	if ((smf_flags & MDL_PITCHFROMMOMENTUM))
 	{
 		const double x = actor->Vel.X;
 		const double y = actor->Vel.Y;
@@ -101,7 +107,7 @@ void RenderModel(FModelRenderer *renderer, float x, float y, float z, FSpriteMod
 		}
 	}
 
-	if (smf->flags & MDL_ROTATING)
+	if (smf_flags & MDL_ROTATING)
 	{
 		if (smf->rotationSpeed > 0.0000000001 || smf->rotationSpeed < -0.0000000001)
 		{
@@ -117,13 +123,13 @@ void RenderModel(FModelRenderer *renderer, float x, float y, float z, FSpriteMod
 
 	// Added MDL_USEACTORPITCH and MDL_USEACTORROLL flags processing.
 	// If both flags MDL_USEACTORPITCH and MDL_PITCHFROMMOMENTUM are set, the pitch sums up the actor pitch and the velocity vector pitch.
-	if (smf->flags & MDL_USEACTORPITCH)
+	if (smf_flags & MDL_USEACTORPITCH)
 	{
-		double d = angles.Pitch.Degrees;
-		if (smf->flags & MDL_BADROTATION) pitch += d;
+		double d = angles.Pitch.Degrees();
+		if (smf_flags & MDL_BADROTATION) pitch += d;
 		else pitch -= d;
 	}
-	if (smf->flags & MDL_USEACTORROLL) roll += angles.Roll.Degrees;
+	if (smf_flags & MDL_USEACTORROLL) roll += angles.Roll.Degrees();
 
 	VSMatrix objectToWorldMatrix;
 	objectToWorldMatrix.loadIdentity();
@@ -132,58 +138,72 @@ void RenderModel(FModelRenderer *renderer, float x, float y, float z, FSpriteMod
 	objectToWorldMatrix.translate(x, z, y);
 
 	// [Nash] take SpriteRotation into account
-	angle += actor->SpriteRotation.Degrees;
+	angle += actor->SpriteRotation.Degrees();
+
+	// consider the pixel stretching. For non-voxels this must be factored out here
+	float stretch = 1.f;
+
+	// [MK] distortions might happen depending on when the pixel stretch is compensated for
+	// so we make the "undistorted" behavior opt-in
+	if ((smf_flags & MDL_CORRECTPIXELSTRETCH) && smf->modelIDs.Size() > 0)
+	{
+		stretch = (smf->modelIDs[0] >= 0 ? Models[smf->modelIDs[0]]->getAspectFactor(actor->Level->info->pixelstretch) : 1.f) / actor->Level->info->pixelstretch;
+		objectToWorldMatrix.scale(1, stretch, 1);
+	}
 
 	// Applying model transformations:
 	// 1) Applying actor angle, pitch and roll to the model
-	if (smf->flags & MDL_USEROTATIONCENTER)
+	if (smf_flags & MDL_USEROTATIONCENTER)
 	{
-		objectToWorldMatrix.translate(smf->rotationCenterX, smf->rotationCenterZ, smf->rotationCenterY);
+		objectToWorldMatrix.translate(smf->rotationCenterX, smf->rotationCenterZ/stretch, smf->rotationCenterY);
 	}
 	objectToWorldMatrix.rotate(-angle, 0, 1, 0);
 	objectToWorldMatrix.rotate(pitch, 0, 0, 1);
 	objectToWorldMatrix.rotate(-roll, 1, 0, 0);
-	if (smf->flags & MDL_USEROTATIONCENTER)
+	if (smf_flags & MDL_USEROTATIONCENTER)
 	{
-		objectToWorldMatrix.translate(-smf->rotationCenterX, -smf->rotationCenterZ, -smf->rotationCenterY);
+		objectToWorldMatrix.translate(-smf->rotationCenterX, -smf->rotationCenterZ/stretch, -smf->rotationCenterY);
 	}
 
 	// 2) Applying Doomsday like rotation of the weapon pickup models
 	// The rotation angle is based on the elapsed time.
 
-	if (smf->flags & MDL_ROTATING)
+	if (smf_flags & MDL_ROTATING)
 	{
-		objectToWorldMatrix.translate(smf->rotationCenterX, smf->rotationCenterY, smf->rotationCenterZ);
+		objectToWorldMatrix.translate(smf->rotationCenterX, smf->rotationCenterY/stretch, smf->rotationCenterZ);
 		objectToWorldMatrix.rotate(rotateOffset, smf->xrotate, smf->yrotate, smf->zrotate);
-		objectToWorldMatrix.translate(-smf->rotationCenterX, -smf->rotationCenterY, -smf->rotationCenterZ);
+		objectToWorldMatrix.translate(-smf->rotationCenterX, -smf->rotationCenterY/stretch, -smf->rotationCenterZ);
 	}
 
 	// 3) Scaling model.
 	objectToWorldMatrix.scale(scaleFactorX, scaleFactorZ, scaleFactorY);
 
 	// 4) Aplying model offsets (model offsets do not depend on model scalings).
-	objectToWorldMatrix.translate(smf->xoffset / smf->xscale, smf->zoffset / smf->zscale, smf->yoffset / smf->yscale);
+	objectToWorldMatrix.translate(smf->xoffset / smf->xscale, smf->zoffset / (smf->zscale*stretch), smf->yoffset / smf->yscale);
 
 	// 5) Applying model rotations.
 	objectToWorldMatrix.rotate(-smf->angleoffset, 0, 1, 0);
 	objectToWorldMatrix.rotate(smf->pitchoffset, 0, 0, 1);
 	objectToWorldMatrix.rotate(-smf->rolloffset, 1, 0, 0);
 
-	// consider the pixel stretching. For non-voxels this must be factored out here
-	float stretch = (smf->modelIDs[0] != -1 ? Models[smf->modelIDs[0]]->getAspectFactor(actor->Level->info->pixelstretch) : 1.f) / actor->Level->info->pixelstretch;
-	objectToWorldMatrix.scale(1, stretch, 1);
+	if (!(smf_flags & MDL_CORRECTPIXELSTRETCH) && smf->modelIDs.Size() > 0)
+	{
+		stretch = (smf->modelIDs[0] >= 0 ? Models[smf->modelIDs[0]]->getAspectFactor(actor->Level->info->pixelstretch) : 1.f) / actor->Level->info->pixelstretch;
+		objectToWorldMatrix.scale(1, stretch, 1);
+	}
 
 	float orientation = scaleFactorX * scaleFactorY * scaleFactorZ;
 
-	renderer->BeginDrawModel(actor->RenderStyle, smf, objectToWorldMatrix, orientation < 0);
-	RenderFrameModels(renderer, actor->Level, smf, actor->state, actor->tics, actor->GetClass(), translation);
-	renderer->EndDrawModel(actor->RenderStyle, smf);
+	renderer->BeginDrawModel(actor->RenderStyle, smf_flags, objectToWorldMatrix, orientation < 0);
+	RenderFrameModels(renderer, actor->Level, smf, actor->state, actor->tics, translation, actor);
+	renderer->EndDrawModel(actor->RenderStyle, smf_flags);
 }
 
-void RenderHUDModel(FModelRenderer *renderer, DPSprite *psp, float ofsX, float ofsY)
+void RenderHUDModel(FModelRenderer *renderer, DPSprite *psp, FVector3 translation, FVector3 rotation, FVector3 rotation_pivot, FSpriteModelFrame *smf)
 {
 	AActor * playermo = players[consoleplayer].camera;
-	FSpriteModelFrame *smf = psp->Caller != nullptr ? FindModelFrame(psp->Caller->GetClass(), psp->GetSprite(), psp->GetFrame(), false) : nullptr;
+
+	int smf_flags = smf->getFlags(psp->Caller->modelData);
 
 	// [BB] No model found for this sprite, so we can't render anything.
 	if (smf == nullptr)
@@ -193,15 +213,34 @@ void RenderHUDModel(FModelRenderer *renderer, DPSprite *psp, float ofsX, float o
 	// but we need to position it correctly in the world for light to work properly.
 	VSMatrix objectToWorldMatrix = renderer->GetViewToWorldMatrix();
 
+	// [Nash] Optional scale weapon FOV
+	float fovscale = 1.0f;
+	if (smf_flags & MDL_SCALEWEAPONFOV)
+	{
+		fovscale = tan(players[consoleplayer].DesiredFOV * (0.5f * M_PI / 180.f));
+		fovscale = 1.f + (fovscale - 1.f) * cl_scaleweaponfov;
+	}
+
 	// Scaling model (y scale for a sprite means height, i.e. z in the world!).
-	objectToWorldMatrix.scale(smf->xscale, smf->zscale, smf->yscale);
+	objectToWorldMatrix.scale(smf->xscale, smf->zscale, smf->yscale / fovscale);
 
 	// Aplying model offsets (model offsets do not depend on model scalings).
 	objectToWorldMatrix.translate(smf->xoffset / smf->xscale, smf->zoffset / smf->zscale, smf->yoffset / smf->yscale);
 
 	// [BB] Weapon bob, very similar to the normal Doom weapon bob.
-	objectToWorldMatrix.rotate(ofsX / 4, 0, 1, 0);
-	objectToWorldMatrix.rotate((ofsY - WEAPONTOP) / -4., 1, 0, 0);
+
+	
+
+	objectToWorldMatrix.translate(rotation_pivot.X, rotation_pivot.Y, rotation_pivot.Z);
+	
+	objectToWorldMatrix.rotate(rotation.X, 0, 1, 0);
+	objectToWorldMatrix.rotate(rotation.Y, 1, 0, 0);
+	objectToWorldMatrix.rotate(rotation.Z, 0, 0, 1);
+
+	objectToWorldMatrix.translate(-rotation_pivot.X, -rotation_pivot.Y, -rotation_pivot.Z);
+	
+	objectToWorldMatrix.translate(translation.X, translation.Y, translation.Z);
+	
 
 	// [BB] For some reason the jDoom models need to be rotated.
 	objectToWorldMatrix.rotate(90.f, 0, 1, 0);
@@ -213,20 +252,93 @@ void RenderHUDModel(FModelRenderer *renderer, DPSprite *psp, float ofsX, float o
 
 	float orientation = smf->xscale * smf->yscale * smf->zscale;
 
-	renderer->BeginDrawHUDModel(playermo->RenderStyle, objectToWorldMatrix, orientation < 0);
-	uint32_t trans = psp->GetTranslation() != 0 ? psp->GetTranslation() : 0;
+	renderer->BeginDrawHUDModel(playermo->RenderStyle, objectToWorldMatrix, orientation < 0, smf_flags);
+	auto trans = psp->GetTranslation();
 	if ((psp->Flags & PSPF_PLAYERTRANSLATED)) trans = psp->Owner->mo->Translation;
-	RenderFrameModels(renderer, playermo->Level, smf, psp->GetState(), psp->GetTics(), psp->Caller->GetClass(), trans);
-	renderer->EndDrawHUDModel(playermo->RenderStyle);
+
+	RenderFrameModels(renderer, playermo->Level, smf, psp->GetState(), psp->GetTics(), trans, psp->Caller);
+	renderer->EndDrawHUDModel(playermo->RenderStyle, smf_flags);
 }
 
-void RenderFrameModels(FModelRenderer *renderer, FLevelLocals *Level, const FSpriteModelFrame *smf, const FState *curState, const int curTics, const PClass *ti, int translation)
+double getCurrentFrame(const ModelAnim &anim, double tic, bool *looped)
+{
+	if(anim.framerate <= 0) return anim.startFrame;
+
+	double frame = ((tic - anim.startTic) / GameTicRate) * anim.framerate; // position in frames
+
+	double duration = double(anim.lastFrame) - anim.startFrame;
+
+	if((anim.flags & MODELANIM_LOOP) && frame >= duration)
+	{
+		if(looped) *looped = true;
+		frame = frame - duration;
+		return fmod(frame, anim.lastFrame - anim.loopFrame) + anim.loopFrame;
+	}
+	else
+	{
+		return min(frame, duration) + anim.startFrame;
+	}
+}
+
+void calcFrame(const ModelAnim &anim, double tic, ModelAnimFrameInterp &inter)
+{
+	bool looped = false;
+
+	double frame = getCurrentFrame(anim, tic, &looped);
+
+	inter.frame1 = int(floor(frame));
+
+	inter.inter = frame - inter.frame1;
+
+	inter.frame2 = int(ceil(frame));
+
+	int startFrame = (looped ? anim.loopFrame : anim.startFrame);
+
+	if(inter.frame1 < startFrame) inter.frame1 = anim.lastFrame;
+	if(inter.frame2 > anim.lastFrame) inter.frame2 = startFrame;
+}
+
+void calcFrames(const ModelAnim &curAnim, double tic, ModelAnimFrameInterp &to, float &inter)
+{
+	if(curAnim.startTic > tic)
+	{
+		inter = (tic - (curAnim.startTic - curAnim.switchOffset)) / curAnim.switchOffset;
+
+		calcFrame(curAnim, curAnim.startTic, to);
+	}
+	else
+	{
+		inter = -1.0f;
+		calcFrame(curAnim, tic, to);
+	}
+}
+
+CalcModelFrameInfo CalcModelFrame(FLevelLocals *Level, const FSpriteModelFrame *smf, const FState *curState, const int curTics, DActorModelData* data, AActor* actor, bool is_decoupled, double tic)
 {
 	// [BB] Frame interpolation: Find the FSpriteModelFrame smfNext which follows after smf in the animation
 	// and the scalar value inter ( element of [0,1) ), both necessary to determine the interpolated frame.
-	FSpriteModelFrame * smfNext = nullptr;
-	double inter = 0.;
-	if (gl_interpolate_model_frames && !(smf->flags & MDL_NOINTERPOLATION))
+
+	int smf_flags = smf->getFlags(data);
+
+	const FSpriteModelFrame * smfNext = nullptr;
+	float inter = 0.;
+
+	ModelAnimFrameInterp decoupled_frame;
+	ModelAnimFrame * decoupled_frame_prev = nullptr;
+
+	// if prev_frame == -1: interpolate(main_frame, next_frame, inter), else: interpolate(interpolate(main_prev_frame, main_frame, inter_main), interpolate(next_prev_frame, next_frame, inter_next), inter)
+	// 4-way interpolation is needed to interpolate animation switches between animations that aren't 35hz
+
+	if(is_decoupled)
+	{
+		smfNext = smf = &BaseSpriteModelFrames[actor->GetClass()];
+		if(data && !(data->curAnim.flags & MODELANIM_NONE))
+		{
+			calcFrames(data->curAnim, tic, decoupled_frame, inter);
+			decoupled_frame_prev = &data->prevAnim;
+		}
+	}
+	else if (gl_interpolate_model_frames && !(smf_flags & MDL_NOINTERPOLATION))
 	{
 		FState *nextState = curState->GetNextState();
 		if (curState != nextState && nextState)
@@ -234,7 +346,7 @@ void RenderFrameModels(FModelRenderer *renderer, FLevelLocals *Level, const FSpr
 			// [BB] To interpolate at more than 35 fps we take tic fractions into account.
 			float ticFraction = 0.;
 			// [BB] In case the tic counter is frozen we have to leave ticFraction at zero.
-			if (ConsoleState == c_up && menuactive != MENU_On && !Level->isFrozen())
+			if ((ConsoleState == c_up || ConsoleState == c_rising) && (menuactive == MENU_Off || menuactive == MENU_OnNoPause) && !Level->isFrozen())
 			{
 				ticFraction = I_GetTimeFrac();
 			}
@@ -248,7 +360,7 @@ void RenderFrameModels(FModelRenderer *renderer, FLevelLocals *Level, const FSpr
 			{
 				// [BB] Workaround for actors that use the same frame twice in a row.
 				// Most of the standard Doom monsters do this in their see state.
-				if ((smf->flags & MDL_INTERPOLATEDOUBLEDFRAMES))
+				if ((smf_flags & MDL_INTERPOLATEDOUBLEDFRAMES))
 				{
 					const FState *prevState = curState - 1;
 					if ((curState->sprite == prevState->sprite) && (curState->Frame == prevState->Frame))
@@ -263,25 +375,233 @@ void RenderFrameModels(FModelRenderer *renderer, FLevelLocals *Level, const FSpr
 					}
 				}
 				if (nextState && inter != 0.0)
-					smfNext = FindModelFrame(ti, nextState->sprite, nextState->Frame, false);
+					smfNext = FindModelFrame(actor, nextState->sprite, nextState->Frame, false);
 			}
 		}
 	}
 
-	for (int i = 0; i < smf->modelsAmount; i++)
+	unsigned modelsamount = smf->modelsAmount;
+	//[SM] - if we added any models for the frame to also render, then we also need to update modelsAmount for this smf
+	if (data != nullptr)
 	{
-		if (smf->modelIDs[i] != -1)
+		if (data->models.Size() > modelsamount)
+			modelsamount = data->models.Size();
+	}
+
+	return
+	{
+		smf_flags,
+		smfNext,
+		inter,
+		is_decoupled,
+		decoupled_frame,
+		decoupled_frame_prev,
+		modelsamount
+	};
+}
+
+bool CalcModelOverrides(int i, const FSpriteModelFrame *smf, DActorModelData* data, const CalcModelFrameInfo &info, ModelDrawInfo &out, bool is_decoupled)
+{
+	//reset drawinfo
+	out.modelid = -1;
+	out.animationid = -1;
+	out.modelframe = -1;
+	out.modelframenext = -1;
+	out.skinid.SetNull();
+	out.surfaceskinids.Clear();
+
+	if (data)
+	{
+		//modelID
+		if (data->models.Size() > i && data->models[i].modelID >= 0)
 		{
-			FModel * mdl = Models[smf->modelIDs[i]];
-			auto tex = smf->skinIDs[i].isValid() ? TexMan.GetGameTexture(smf->skinIDs[i], true) : nullptr;
-			mdl->BuildVertexBuffer(renderer);
+			out.modelid = data->models[i].modelID;
+		}
+		else if(data->models.Size() > i && data->models[i].modelID == -2)
+		{
+			return false;
+		}
+		else if(smf->modelsAmount > i)
+		{
+			out.modelid = smf->modelIDs[i];
+		}
 
-			mdl->PushSpriteMDLFrame(smf, i);
+		//animationID
+		if (data->animationIDs.Size() > i && data->animationIDs[i] >= 0)
+		{
+			out.animationid = data->animationIDs[i];
+		}
+		else if(smf->modelsAmount > i)
+		{
+			out.animationid = smf->animationIDs[i];
+		}
+		if(!is_decoupled)
+		{
+			//modelFrame
+			if (data->modelFrameGenerators.Size() > i
+				&& (unsigned)data->modelFrameGenerators[i] < info.modelsamount
+				&& smf->modelframes[data->modelFrameGenerators[i]] >= 0
+				) {
+				out.modelframe = smf->modelframes[data->modelFrameGenerators[i]];
 
-			if (smfNext && smf->modelframes[i] != smfNext->modelframes[i])
-				mdl->RenderFrame(renderer, tex, smf->modelframes[i], smfNext->modelframes[i], inter, translation);
-			else
-				mdl->RenderFrame(renderer, tex, smf->modelframes[i], smf->modelframes[i], 0.f, translation);
+				if (info.smfNext) 
+				{
+					if(info.smfNext->modelframes[data->modelFrameGenerators[i]] >= 0)
+					{
+						out.modelframenext = info.smfNext->modelframes[data->modelFrameGenerators[i]];
+					}
+					else
+					{
+						out.modelframenext = info.smfNext->modelframes[i];
+					}
+				}
+			}
+			else if(smf->modelsAmount > i)
+			{
+				out.modelframe = smf->modelframes[i];
+				if (info.smfNext) out.modelframenext = info.smfNext->modelframes[i];
+			}
+		}
+
+		//skinID
+		if (data->skinIDs.Size() > i && data->skinIDs[i].isValid())
+		{
+			out.skinid = data->skinIDs[i];
+		}
+		else if(smf->modelsAmount > i)
+		{
+			out.skinid = smf->skinIDs[i];
+		}
+
+		//surfaceSkinIDs
+		if(data->models.Size() > i && data->models[i].surfaceSkinIDs.Size() > 0)
+		{
+			unsigned sz1 = smf->surfaceskinIDs.Size();
+			unsigned sz2 = data->models[i].surfaceSkinIDs.Size();
+			unsigned start = i * MD3_MAX_SURFACES;
+
+			out.surfaceskinids = data->models[i].surfaceSkinIDs;
+			out.surfaceskinids.Resize(MD3_MAX_SURFACES);
+
+			for (unsigned surface = 0; surface < MD3_MAX_SURFACES; surface++)
+			{
+				if (sz2 > surface && (data->models[i].surfaceSkinIDs[surface].isValid()))
+				{
+					continue;
+				}
+				if((surface + start) < sz1)
+				{
+					out.surfaceskinids[surface] = smf->surfaceskinIDs[surface + start];
+				}
+				else
+				{
+					out.surfaceskinids[surface].SetNull();
+				}
+			}
+		}
+	}
+	else
+	{
+		out.modelid = smf->modelIDs[i];
+		out.animationid = smf->animationIDs[i];
+		out.modelframe = smf->modelframes[i];
+		if (info.smfNext) out.modelframenext = info.smfNext->modelframes[i];
+		out.skinid = smf->skinIDs[i];
+	}
+
+	return (out.modelid >= 0 && out.modelid < Models.size());
+}
+
+
+const TArray<VSMatrix> * ProcessModelFrame(FModel * animation, bool nextFrame, int i, const FSpriteModelFrame *smf, DActorModelData* modelData, const CalcModelFrameInfo &frameinfo, ModelDrawInfo &drawinfo, bool is_decoupled, double tic)
+{
+	const TArray<TRS>* animationData = nullptr;
+	
+	if (drawinfo.animationid >= 0)
+	{
+		animation = Models[drawinfo.animationid];
+		animationData = animation->AttachAnimationData();
+	}
+
+	const TArray<VSMatrix> *boneData = nullptr;
+
+	if(is_decoupled)
+	{
+		if(frameinfo.decoupled_frame.frame1 >= 0)
+		{
+			boneData = animation->CalculateBones(
+				frameinfo.decoupled_frame_prev ? *frameinfo.decoupled_frame_prev : nullptr,
+				frameinfo.decoupled_frame,
+				frameinfo.inter,
+				animationData);
+		}
+	}
+	else
+	{
+		boneData = animation->CalculateBones(
+			nullptr,
+			{
+				nextFrame ? frameinfo.inter : -1.0f,
+				drawinfo.modelframe,
+				drawinfo.modelframenext
+			},
+			-1.0f,
+			animationData);
+	}
+
+	return boneData;
+}
+
+static inline void RenderModelFrame(FModelRenderer *renderer, int i, const FSpriteModelFrame *smf, DActorModelData* modelData, const CalcModelFrameInfo &frameinfo, ModelDrawInfo &drawinfo, bool is_decoupled, double tic, FTranslationID translation, int &boneStartingPosition, bool &evaluatedSingle)
+{
+	FModel * mdl = Models[drawinfo.modelid];
+	auto tex = drawinfo.skinid.isValid() ? TexMan.GetGameTexture(drawinfo.skinid, true) : nullptr;
+	mdl->BuildVertexBuffer(renderer);
+
+	auto ssidp = drawinfo.surfaceskinids.Size() > 0
+		? drawinfo.surfaceskinids.Data()
+		: (((i * MD3_MAX_SURFACES) < smf->surfaceskinIDs.Size()) ? &smf->surfaceskinIDs[i * MD3_MAX_SURFACES] : nullptr);
+
+	bool nextFrame = frameinfo.smfNext && drawinfo.modelframe != drawinfo.modelframenext;
+
+	// [Jay] while per-model animations aren't done, DECOUPLEDANIMATIONS does the same as MODELSAREATTACHMENTS
+	if(!evaluatedSingle)
+	{  // [Jay] TODO per-model decoupled animations
+		const TArray<VSMatrix> *boneData = ProcessModelFrame(mdl, nextFrame, i, smf, modelData, frameinfo, drawinfo, is_decoupled, tic);
+
+		if(frameinfo.smf_flags & MDL_MODELSAREATTACHMENTS || is_decoupled)
+		{
+			boneStartingPosition = boneData ? screen->mBones->UploadBones(*boneData) : -1;
+			evaluatedSingle = true;
+		}
+	}
+
+	mdl->RenderFrame(renderer, tex, drawinfo.modelframe, nextFrame ? drawinfo.modelframenext : drawinfo.modelframe, nextFrame ? frameinfo.inter : -1.f, translation, ssidp, boneStartingPosition);
+}
+
+void RenderFrameModels(FModelRenderer *renderer, FLevelLocals *Level, const FSpriteModelFrame *smf, const FState *curState, const int curTics, FTranslationID translation, AActor* actor)
+{
+	double tic = actor->Level->totaltime;
+	if ((ConsoleState == c_up || ConsoleState == c_rising) && (menuactive == MENU_Off || menuactive == MENU_OnNoPause) && !actor->isFrozen())
+	{
+		tic += I_GetTimeFrac();
+	}
+
+	bool is_decoupled = (actor->flags9 & MF9_DECOUPLEDANIMATIONS);
+
+	DActorModelData* modelData = actor ? actor->modelData.ForceGet() : nullptr;
+
+	CalcModelFrameInfo frameinfo = CalcModelFrame(Level, smf, curState, curTics, modelData, actor, is_decoupled, tic);
+	ModelDrawInfo drawinfo;
+
+	int boneStartingPosition = -1;
+	bool evaluatedSingle = false;
+
+	for (unsigned i = 0; i < frameinfo.modelsamount; i++)
+	{
+		if (CalcModelOverrides(i, smf, modelData, frameinfo, drawinfo, is_decoupled))
+		{
+			RenderModelFrame(renderer, i, smf, modelData, frameinfo, drawinfo, is_decoupled, tic, translation, boneStartingPosition, evaluatedSingle);
 		}
 	}
 }
@@ -296,7 +616,7 @@ static TArray<int> SpriteModelHash;
 //
 //===========================================================================
 
-static void ParseModelDefLump(int Lump);
+void ParseModelDefLump(int Lump);
 
 void InitModels()
 {
@@ -324,8 +644,18 @@ void InitModels()
 		smf.modelIDs[0] = VoxelDefs[i]->Voxel->VoxelIndex;
 		smf.skinIDs.Alloc(1);
 		smf.skinIDs[0] = md->GetPaletteTexture();
+		smf.animationIDs.Alloc(1);
+		smf.animationIDs[0] = -1;
 		smf.xscale = smf.yscale = smf.zscale = VoxelDefs[i]->Scale;
-		smf.angleoffset = VoxelDefs[i]->AngleOffset.Degrees;
+		smf.angleoffset = VoxelDefs[i]->AngleOffset.Degrees();
+		smf.xoffset = VoxelDefs[i]->xoffset;
+		smf.yoffset = VoxelDefs[i]->yoffset;
+		smf.zoffset = VoxelDefs[i]->zoffset;
+		// this helps catching uninitialized data.
+		assert(VoxelDefs[i]->PitchFromMomentum == true || VoxelDefs[i]->PitchFromMomentum == false);
+		if (VoxelDefs[i]->PitchFromMomentum) smf.flags |= MDL_PITCHFROMMOMENTUM;
+		if (VoxelDefs[i]->UseActorPitch) smf.flags |= MDL_USEACTORPITCH;
+		if (VoxelDefs[i]->UseActorRoll) smf.flags |= MDL_USEACTORROLL;
 		if (VoxelDefs[i]->PlacedSpin != 0)
 		{
 			smf.yrotate = 1.f;
@@ -371,7 +701,7 @@ void InitModels()
 	}
 }
 
-static void ParseModelDefLump(int Lump)
+void ParseModelDefLump(int Lump)
 {
 	FScanner sc(Lump);
 	while (sc.GetString())
@@ -423,6 +753,7 @@ static void ParseModelDefLump(int Lump)
 			initArray(smf.modelIDs, smf.modelsAmount, -1);
 			initArray(smf.skinIDs, smf.modelsAmount, FNullTextureID());
 			initArray(smf.surfaceskinIDs, smf.modelsAmount * MD3_MAX_SURFACES, FNullTextureID());
+			initArray(smf.animationIDs, smf.modelsAmount, -1);
 			initArray(smf.modelframes, smf.modelsAmount, 0);
 
 			sc.RestorePos(scPos);
@@ -457,6 +788,26 @@ static void ParseModelDefLump(int Lump)
 						Printf("%s: model not found in %s\n", sc.String, path.GetChars());
 					}
 				}
+				else if (sc.Compare("animation"))
+				{
+					sc.MustGetNumber();
+					index = sc.Number;
+					if (index < 0)
+					{
+						sc.ScriptError("Animation index must be 0 or greater in %s", type->TypeName.GetChars());
+					}
+					else if (index >= smf.modelsAmount)
+					{
+						sc.ScriptError("Too many models in %s", type->TypeName.GetChars());
+					}
+					sc.MustGetString();
+					FixPathSeperator(sc.String);
+					smf.animationIDs[index] = FindModel(path.GetChars(), sc.String);
+					if (smf.animationIDs[index] == -1)
+					{
+						Printf("%s: animation model not found in %s\n", sc.String, path.GetChars());
+					}
+				}
 				else if (sc.Compare("scale"))
 				{
 					sc.MustGetFloat();
@@ -524,6 +875,18 @@ static void ParseModelDefLump(int Lump)
 				{
 					smf.flags |= MDL_USEACTORROLL;
 				}
+				else if (sc.Compare("noperpixellighting"))
+				{
+					smf.flags |= MDL_NOPERPIXELLIGHTING;
+				}
+				else if (sc.Compare("scaleweaponfov"))
+				{
+					smf.flags |= MDL_SCALEWEAPONFOV;
+				}
+				else if (sc.Compare("modelsareattachments"))
+				{
+					smf.flags |= MDL_MODELSAREATTACHMENTS;
+				}
 				else if (sc.Compare("rotating"))
 				{
 					smf.flags |= MDL_ROTATING;
@@ -624,6 +987,16 @@ static void ParseModelDefLump(int Lump)
 						}
 					}
 				}
+				else if (sc.Compare("baseframe"))
+				{
+					FSpriteModelFrame *smfp = &BaseSpriteModelFrames.Insert(type, smf);
+					for(int modelID : smf.modelIDs)
+					{
+						if(modelID >= 0)
+							Models[modelID]->baseFrame = smfp;
+					}
+					GetDefaultByType(type)->hasmodel = true;
+				}
 				else if (sc.Compare("frameindex") || sc.Compare("frame"))
 				{
 					bool isframe=!!sc.Compare("frame");
@@ -659,9 +1032,13 @@ static void ParseModelDefLump(int Lump)
 					if (isframe)
 					{
 						sc.MustGetString();
-						if (smf.modelIDs[index] != -1)
+						if (smf.modelIDs[index] >= 0)
 						{
 							FModel *model = Models[smf.modelIDs[index]];
+							if (smf.animationIDs[index] >= 0)
+							{
+								model = Models[smf.animationIDs[index]];
+							}
 							smf.modelframes[index] = model->FindFrame(sc.String);
 							if (smf.modelframes[index]==-1) sc.ScriptError("Unknown frame '%s' in %s", sc.String, type->TypeName.GetChars());
 						}
@@ -700,6 +1077,14 @@ static void ParseModelDefLump(int Lump)
 					smf.rotationCenterY = 0.;
 					smf.rotationCenterZ = 0.;
 				}
+				else if (sc.Compare("correctpixelstretch"))
+				{
+					smf.flags |= MDL_CORRECTPIXELSTRETCH;
+				}
+				else if (sc.Compare("forcecullbackfaces"))
+				{
+					smf.flags |= MDL_FORCECULLBACKFACES;
+				}
 				else
 				{
 					sc.ScriptMessage("Unrecognized string \"%s\"", sc.String);
@@ -730,24 +1115,24 @@ static void ParseModelDefLump(int Lump)
 //
 //===========================================================================
 
-FSpriteModelFrame * FindModelFrame(const PClass * ti, int sprite, int frame, bool dropped)
+FSpriteModelFrame * FindModelFrameRaw(const AActor * actorDefaults, const PClass * ti, int sprite, int frame, bool dropped)
 {
-	if (GetDefaultByType(ti)->hasmodel)
+	if(actorDefaults->hasmodel)
 	{
 		FSpriteModelFrame smf;
 
 		memset(&smf, 0, sizeof(smf));
-		smf.type=ti;
-		smf.sprite=sprite;
-		smf.frame=frame;
+		smf.type = ti;
+		smf.sprite = sprite;
+		smf.frame = frame;
 
 		int hash = SpriteModelHash[ModelFrameHash(&smf) % SpriteModelFrames.Size()];
 
 		while (hash>=0)
 		{
 			FSpriteModelFrame * smff = &SpriteModelFrames[hash];
-			if (smff->type==ti && smff->sprite==sprite && smff->frame==frame) return smff;
-			hash=smff->hashnext;
+			if (smff->type == ti && smff->sprite == sprite && smff->frame == frame) return smff;
+			hash = smff->hashnext;
 		}
 	}
 
@@ -761,14 +1146,52 @@ FSpriteModelFrame * FindModelFrame(const PClass * ti, int sprite, int frame, boo
 			if (sprframe->Voxel != nullptr)
 			{
 				int index = sprframe->Voxel->VoxeldefIndex;
-				if (dropped && sprframe->Voxel->DroppedSpin !=sprframe->Voxel->PlacedSpin) index++;
+				if (dropped && sprframe->Voxel->DroppedSpin != sprframe->Voxel->PlacedSpin) index++;
 				return &SpriteModelFrames[index];
 			}
 		}
 	}
+
 	return nullptr;
 }
 
+FSpriteModelFrame * FindModelFrame(const PClass * ti, int sprite, int frame, bool dropped)
+{
+	auto def = GetDefaultByType(ti);
+
+	if (def->hasmodel)
+	{
+		if(def->flags9 & MF9_DECOUPLEDANIMATIONS)
+		{
+			FSpriteModelFrame * smf = BaseSpriteModelFrames.CheckKey(ti);
+			if(smf) return smf;
+		}
+	}
+
+	return FindModelFrameRaw(def, ti, sprite, frame, dropped);
+}
+
+FSpriteModelFrame * FindModelFrame(const PClass * ti, bool is_decoupled, int sprite, int frame, bool dropped)
+{
+	if(!ti) return nullptr;
+
+	if(is_decoupled)
+	{
+		return BaseSpriteModelFrames.CheckKey(ti);
+	}
+	else
+	{
+		return FindModelFrameRaw(GetDefaultByType(ti), ti, sprite, frame, dropped);
+	}
+}
+
+FSpriteModelFrame * FindModelFrame(AActor * thing, int sprite, int frame, bool dropped)
+{
+	if(!thing) return nullptr;
+
+	return FindModelFrame((thing->modelData != nullptr && thing->modelData->modelDef != nullptr) ? thing->modelData->modelDef : thing->GetClass(), (thing->flags9 & MF9_DECOUPLEDANIMATIONS), sprite, frame, dropped);
+}
+
 //===========================================================================
 //
 // IsHUDModelForPlayerAvailable
@@ -783,8 +1206,13 @@ bool IsHUDModelForPlayerAvailable (player_t * player)
 	// [MK] check that at least one psprite uses models
 	for (DPSprite *psp = player->psprites; psp != nullptr && psp->GetID() < PSP_TARGETCENTER; psp = psp->GetNext())
 	{
-		FSpriteModelFrame *smf = psp->Caller != nullptr ? FindModelFrame(psp->Caller->GetClass(), psp->GetSprite(), psp->GetFrame(), false) : nullptr;
-		if ( smf != nullptr ) return true;
+		if ( FindModelFrame(psp->Caller, psp->GetSprite(), psp->GetFrame(), false) != nullptr ) return true;
 	}
 	return false;
 }
+
+
+unsigned int FSpriteModelFrame::getFlags(class DActorModelData * defs) const
+{
+	return (defs && defs->flags & MODELDATA_OVERRIDE_FLAGS)? (flags | defs->overrideFlagsSet) & ~(defs->overrideFlagsClear) : flags;
+}
\ No newline at end of file
