@@ -42,6 +42,9 @@
 #include "hw_clock.h"
 #include "flatvertices.h"
 #include "hw_vertexbuilder.h"
+#include "hw_walldispatcher.h"
+
+#include "p_visualthinker.h"
 
 #ifdef ARCH_IA32
 #include <immintrin.h>
@@ -50,11 +53,24 @@
 CVAR(Bool, gl_multithread, true, CVAR_ARCHIVE | CVAR_GLOBALCONFIG)
 
 EXTERN_CVAR(Float, r_actorspriteshadowdist)
+EXTERN_CVAR(Bool, r_radarclipper)
+EXTERN_CVAR(Bool, r_dithertransparency)
 
 thread_local bool isWorkerThread;
 ctpl::thread_pool renderPool(1);
 bool inited = false;
 
+const int MAXDITHERACTORS = 20; // Maximum number of enemies that can set dither-transparency flags
+AActor* RenderedTargets[MAXDITHERACTORS];
+int RTnum;
+
+void ClearDitherTargets()
+{
+	RTnum = 0; // Number of rendered enemies/targets
+	for (int ii = 0; ii < MAXDITHERACTORS; ii++)
+	  RenderedTargets[ii] = nullptr;
+}
+
 struct RenderJob
 {
 	enum
@@ -105,6 +121,7 @@ static RenderJobQueue jobQueue;	// One static queue is sufficient here. This cod
 void HWDrawInfo::WorkerThread()
 {
 	sector_t *front, *back;
+	HWWallDispatcher disp(this);
 
 	WTTotal.Clock();
 	isWorkerThread = true;	// for adding asserts in GL API code. The worker thread may never call any GL API.
@@ -169,7 +186,7 @@ void HWDrawInfo::WorkerThread()
 			}
 			else back = nullptr;
 
-			wall.Process(this, job->seg, front, back);
+			wall.Process(&disp, job->seg, front, back);
 			rendered_lines++;
 			SetupWall.Unclock();
 			break;
@@ -267,6 +284,28 @@ void HWDrawInfo::AddLine (seg_t *seg, bool portalclip)
 	auto &clipper = *mClipper;
 	angle_t startAngle = clipper.GetClipAngle(seg->v2);
 	angle_t endAngle = clipper.GetClipAngle(seg->v1);
+	auto &clipperr = *rClipper;
+	angle_t startAngleR = 0;
+	angle_t endAngleR = 0;
+	angle_t paddingR = 0x00200000; // Make radar clipping more aggressive (reveal less)
+
+	if(Viewpoint.IsAllowedOoB() && r_radarclipper && !(Level->flags3 & LEVEL3_NOFOGOFWAR))
+	{
+		startAngleR = clipperr.PointToPseudoAngle(seg->v2->fX(), seg->v2->fY());
+		endAngleR = clipperr.PointToPseudoAngle(seg->v1->fX(), seg->v1->fY());
+
+		if (startAngleR - endAngleR >= ANGLE_180)
+		{
+			if (!seg->backsector) clipperr.SafeAddClipRange(startAngleR - paddingR, endAngleR + paddingR);
+			else if((seg->sidedef != nullptr) && !uint8_t(seg->sidedef->Flags & WALLF_POLYOBJ) && (currentsector->sectornum != seg->backsector->sectornum))
+			{
+				if (in_area == area_default) in_area = hw_CheckViewArea(seg->v1, seg->v2, seg->frontsector, seg->backsector);
+				backsector = hw_FakeFlat(seg->backsector, in_area, true);
+				if (hw_CheckClip(seg->sidedef, currentsector, backsector)) clipperr.SafeAddClipRange(startAngleR - paddingR, endAngleR + paddingR);
+				backsector = nullptr;
+			}
+		}
+	}
 
 	// Back side, i.e. backface culling	- read: endAngle >= startAngle!
 	if (startAngle-endAngle<ANGLE_180)  
@@ -278,25 +317,32 @@ void HWDrawInfo::AddLine (seg_t *seg, bool portalclip)
 	{
 		if (!(currentsubsector->flags & SSECMF_DRAWN))
 		{
-			if (clipper.SafeCheckRange(startAngle, endAngle)) 
+			if (clipper.SafeCheckRange(startAngle, endAngle) && (!r_radarclipper || (Level->flags3 & LEVEL3_NOFOGOFWAR)))
 			{
-				currentsubsector->flags |= SSECMF_DRAWN;
+			  currentsubsector->flags |= SSECMF_DRAWN;
+			}
+			if (Viewpoint.IsAllowedOoB() && (r_radarclipper && !(Level->flags3 & LEVEL3_NOFOGOFWAR)) && clipperr.SafeCheckRange(startAngleR, endAngleR))
+			{
+			  currentsubsector->flags |= SSECMF_DRAWN;
 			}
 		}
 		return;
 	}
 
-	if (!clipper.SafeCheckRange(startAngle, endAngle)) 
+	if (!clipper.SafeCheckRange(startAngle, endAngle))
 	{
 		return;
 	}
-	currentsubsector->flags |= SSECMF_DRAWN;
+
+	if (Viewpoint.IsAllowedOoB() && (!r_radarclipper || (Level->flags3 & LEVEL3_NOFOGOFWAR) || clipperr.SafeCheckRange(startAngleR, endAngleR)))
+		currentsubsector->flags |= SSECMF_DRAWN;
 
 	uint8_t ispoly = uint8_t(seg->sidedef->Flags & WALLF_POLYOBJ);
 
 	if (!seg->backsector)
 	{
-		clipper.SafeAddClipRange(startAngle, endAngle);
+		if(!Viewpoint.IsAllowedOoB())
+			if (!(seg->sidedef->Flags & WALLF_DITHERTRANS_MID)) clipper.SafeAddClipRange(startAngle, endAngle);
 	}
 	else if (!ispoly)	// Two-sided polyobjects never obstruct the view
 	{
@@ -323,7 +369,8 @@ void HWDrawInfo::AddLine (seg_t *seg, bool portalclip)
 
 			if (hw_CheckClip(seg->sidedef, currentsector, backsector))
 			{
-				clipper.SafeAddClipRange(startAngle, endAngle);
+				if(!Viewpoint.IsAllowedOoB() && !(seg->sidedef->Flags & WALLF_DITHERTRANS_MID))
+					clipper.SafeAddClipRange(startAngle, endAngle);
 			}
 		}
 	}
@@ -348,9 +395,10 @@ void HWDrawInfo::AddLine (seg_t *seg, bool portalclip)
 			else
 			{
 				HWWall wall;
+				HWWallDispatcher disp(this);
 				SetupWall.Clock();
 				wall.sub = seg->Subsector;
-				wall.Process(this, seg, currentsector, backsector);
+				wall.Process(&disp, seg, currentsector, backsector);
 				rendered_lines++;
 				SetupWall.Unclock();
 			}
@@ -429,11 +477,6 @@ void HWDrawInfo::AddPolyobjs(subsector_t *sub)
 	if (sub->BSP == nullptr || sub->BSP->bDirty)
 	{
 		sub->BuildPolyBSP();
-		for (unsigned i = 0; i < sub->BSP->Segs.Size(); i++)
-		{
-			sub->BSP->Segs[i].Subsector = sub;
-			sub->BSP->Segs[i].PartnerSeg = nullptr;
-		}
 	}
 	if (sub->BSP->Nodes.Size() == 0)
 	{
@@ -490,13 +533,13 @@ void HWDrawInfo::AddLines(subsector_t * sub, sector_t * sector)
 //
 //==========================================================================
 
-inline bool PointOnLine(const DVector2 &pos, const line_t *line)
+inline bool PointOnLine(const DVector2 &pos, const linebase_t *line)
 {
 	double v = (pos.Y - line->v1->fY()) * line->Delta().X + (line->v1->fX() - pos.X) * line->Delta().Y;
 	return fabs(v) <= EQUAL_EPSILON;
 }
 
-void HWDrawInfo::AddSpecialPortalLines(subsector_t * sub, sector_t * sector, line_t *line)
+void HWDrawInfo::AddSpecialPortalLines(subsector_t * sub, sector_t * sector, linebase_t *line)
 {
 	currentsector = sector;
 	currentsubsector = sub;
@@ -528,13 +571,14 @@ void HWDrawInfo::RenderThings(subsector_t * sub, sector_t * sector)
 {
 	sector_t * sec=sub->sector;
 	// Handle all things in sector.
-    const auto &vp = Viewpoint;
+	const auto &vp = Viewpoint;
 	for (auto p = sec->touching_renderthings; p != nullptr; p = p->m_snext)
 	{
 		auto thing = p->m_thing;
 		if (thing->validcount == validcount) continue;
 		thing->validcount = validcount;
 
+		if(Viewpoint.IsAllowedOoB() && thing->Sector->isSecret() && thing->Sector->wasSecret() && !r_radarclipper) continue; // This covers things that are touching non-secret sectors
 		FIntCVar *cvar = thing->GetInfo()->distancecheck;
 		if (cvar != nullptr && *cvar >= 0)
 		{
@@ -599,16 +643,30 @@ void HWDrawInfo::RenderThings(subsector_t * sub, sector_t * sector)
 void HWDrawInfo::RenderParticles(subsector_t *sub, sector_t *front)
 {
 	SetupSprite.Clock();
-	for (int i = Level->ParticlesInSubsec[sub->Index()]; i != NO_PARTICLE; i = Level->Particles[i].snext)
+	for (uint32_t i = 0; i < sub->sprites.Size(); i++)
 	{
+		DVisualThinker *sp = sub->sprites[i];
+		if (!sp || sp->ObjectFlags & OF_EuthanizeMe)
+			continue;
 		if (mClipPortal)
 		{
-			int clipres = mClipPortal->ClipPoint(Level->Particles[i].Pos);
+			int clipres = mClipPortal->ClipPoint(sp->PT.Pos.XY());
 			if (clipres == PClip_InFront) continue;
 		}
 
 		HWSprite sprite;
-		sprite.ProcessParticle(this, &Level->Particles[i], front);
+		sprite.ProcessParticle(this, &sp->PT, front, sp);
+	}
+	for (int i = Level->ParticlesInSubsec[sub->Index()]; i != NO_PARTICLE; i = Level->Particles[i].snext)
+	{
+		if (mClipPortal)
+		{
+			int clipres = mClipPortal->ClipPoint(Level->Particles[i].Pos.XY());
+			if (clipres == PClip_InFront) continue;
+		}
+
+		HWSprite sprite;
+		sprite.ProcessParticle(this, &Level->Particles[i], front, nullptr);
 	}
 	SetupSprite.Unclock();
 }
@@ -653,12 +711,84 @@ void HWDrawInfo::DoSubsector(subsector_t * sub)
 
 	fakesector=hw_FakeFlat(sector, in_area, false);
 
+	if(Viewpoint.IsAllowedOoB() && sector->isSecret() && sector->wasSecret() && !r_radarclipper) return;
+
+	// cull everything if subsector outside all relevant clippers
+	if (Viewpoint.IsAllowedOoB() && (sub->polys == nullptr))
+	{
+		auto &clipper = *mClipper;
+		auto &clipperv = *vClipper;
+		auto &clipperr = *rClipper;
+		int count = sub->numlines;
+		seg_t * seg = sub->firstline;
+		bool anglevisible = false;
+		bool pitchvisible = !(Viewpoint.IsAllowedOoB()); // No vertical clipping if viewpoint is not allowed out of bounds
+		bool radarvisible = !(Viewpoint.IsAllowedOoB()) || !r_radarclipper || (Level->flags3 & LEVEL3_NOFOGOFWAR) || ((sub->flags & SSECMF_DRAWN) && !deathmatch);
+		bool ceilreflect = (mCurrentPortal && strcmp(mCurrentPortal->GetName(), "Planemirror ceiling"));
+		bool floorreflect = (mCurrentPortal && strcmp(mCurrentPortal->GetName(), "Planemirror floor"));
+		double planez = (ceilreflect ? sector->ceilingplane.ZatPoint(Viewpoint.Pos) : sector->floorplane.ZatPoint(Viewpoint.Pos));
+		angle_t pitchtemp;
+		angle_t pitchmin = ANGLE_90;
+		angle_t pitchmax = 0;
+
+		while (count--)
+		{
+			if((seg->v1 != nullptr) && (seg->v2 != nullptr))
+			{
+				if (!anglevisible)
+				{
+					angle_t startAngle = clipper.GetClipAngle(seg->v2);
+					angle_t endAngle = clipper.GetClipAngle(seg->v1);
+					if (startAngle-endAngle >= ANGLE_180) anglevisible |= clipper.SafeCheckRange(startAngle, endAngle);
+				}
+				if (!radarvisible)
+				{
+					angle_t startAngleR = clipperr.PointToPseudoAngle(seg->v2->fX(), seg->v2->fY());
+					angle_t endAngleR = clipperr.PointToPseudoAngle(seg->v1->fX(), seg->v1->fY());
+					if (startAngleR-endAngleR >= ANGLE_180) radarvisible |= clipperr.SafeCheckRange(startAngleR, endAngleR);
+				}
+
+				if (!pitchvisible)
+				{
+					pitchmin = clipperv.PointToPseudoPitch(seg->v1->fX(), seg->v1->fY(),
+														   (ceilreflect || floorreflect) ?
+														   2 * planez - sector->floorplane.ZatPoint(seg->v1) :
+														   sector->floorplane.ZatPoint(seg->v1));
+					pitchmax = clipperv.PointToPseudoPitch(seg->v1->fX(), seg->v1->fY(),
+														   (ceilreflect || floorreflect) ?
+														   2 * planez - sector->ceilingplane.ZatPoint(seg->v1) :
+														   sector->ceilingplane.ZatPoint(seg->v1));
+					pitchvisible |= clipperv.SafeCheckRange(pitchmin, pitchmax);
+				}
+				if (pitchvisible && anglevisible && radarvisible) break;
+				if (!pitchvisible)
+				{
+					pitchtemp = clipperv.PointToPseudoPitch(seg->v2->fX(), seg->v2->fY(),
+															(ceilreflect || floorreflect) ?
+															2 * planez - sector->floorplane.ZatPoint(seg->v2) :
+															sector->floorplane.ZatPoint(seg->v2));
+					if (int(pitchmin) > int(pitchtemp)) pitchmin = pitchtemp;
+					pitchtemp = clipperv.PointToPseudoPitch(seg->v2->fX(), seg->v2->fY(),
+															(ceilreflect || floorreflect) ?
+															2 * planez - sector->ceilingplane.ZatPoint(seg->v2) :
+															sector->ceilingplane.ZatPoint(seg->v2));
+					if (int(pitchmax) < int(pitchtemp)) pitchmax = pitchtemp;
+					pitchvisible |= clipperv.SafeCheckRange(pitchmin, pitchmax);
+				}
+				if (pitchvisible && anglevisible && radarvisible) break;
+			}
+			seg++;
+		}
+		// Skip subsector if outside vertical or horizontal clippers or is in unexplored territory (fog of war)
+		if(!pitchvisible || !anglevisible || (!radarvisible && r_radarclipper)) return;
+	}
+
 	if (mClipPortal)
 	{
 		int clipres = mClipPortal->ClipSubsector(sub);
 		if (clipres == PClip_InFront)
 		{
-			line_t *line = mClipPortal->ClipLine();
+			auto line = mClipPortal->ClipLine();
 			// The subsector is out of range, but we still have to check lines that lie directly on the boundary and may expose their upper or lower parts.
 			if (line) AddSpecialPortalLines(sub, fakesector, line);
 			return;
@@ -671,7 +801,7 @@ void HWDrawInfo::DoSubsector(subsector_t * sub)
 	}
 
 	// [RH] Add particles
-	if (gl_render_things && Level->ParticlesInSubsec[sub->Index()] != NO_PARTICLE)
+	if (gl_render_things && (sub->sprites.Size() > 0 || Level->ParticlesInSubsec[sub->Index()] != NO_PARTICLE))
 	{
 		if (multithread)
 		{
@@ -710,6 +840,20 @@ void HWDrawInfo::DoSubsector(subsector_t * sub)
 				SetupSprite.Unclock();
 			}
 		}
+		if (r_dithertransparency && Viewpoint.IsAllowedOoB() && (RTnum < MAXDITHERACTORS) && mCurrentPortal == nullptr)
+		{
+			// [DVR] Not parallelizable due to variables RTnum and RenderedTargets[]
+			for (auto p = sector->touching_renderthings; p != nullptr; p = p->m_snext)
+			{
+				auto thing = p->m_thing;
+				if (thing->validcount == validcount) continue; // Don't double count
+				if (((thing->flags3 & MF3_ISMONSTER) && !(thing->flags & MF_CORPSE)) || (thing->flags & MF_MISSILE))
+				{
+					if (RTnum < MAXDITHERACTORS) RenderedTargets[RTnum++] = thing;
+					else break;
+				}
+			}
+		}
 	}
 
 	if (gl_render_flats)
@@ -830,19 +974,54 @@ void HWDrawInfo::RenderBSPNode (void *node)
 			if (!(no_renderflags[bsp->Index()] & SSRF_SEEN))
 				return;
 		}
+		if (Viewpoint.IsOrtho())
+		{
+			if (!vClipper->CheckBoxOrthoPitch(bsp->bbox[side]))
+			{
+				if (!(no_renderflags[bsp->Index()] & SSRF_SEEN))
+					return;
+			}
+		}
 
 		node = bsp->children[side];
 	}
 	DoSubsector ((subsector_t *)((uint8_t *)node - 1));
 }
 
+// No need for clipping inside frustum if no fog of war (How is this faster!)
+void HWDrawInfo::RenderOrthoNoFog()
+{
+	if (Viewpoint.IsOrtho() && ((Level->flags3 & LEVEL3_NOFOGOFWAR) || !r_radarclipper))
+	{
+		double vxdbl = Viewpoint.OffPos.X;
+		double vydbl = Viewpoint.OffPos.Y;
+		double ext = Viewpoint.camera->ViewPos->Offset.Length() ?
+			3.0 * Viewpoint.camera->ViewPos->Offset.Length() * tan (Viewpoint.FieldOfView.Radians()*0.5) : 100.0;
+		FBoundingBox viewbox(vxdbl, vydbl, ext);
+
+		for (unsigned int kk = 0; kk < Level->subsectors.Size(); kk++)
+		{
+			if (Level->subsectors[kk].bbox.CheckOverlap(viewbox))
+			{
+				DoSubsector (&Level->subsectors[kk]);
+			}
+		}
+	}
+}
+
 void HWDrawInfo::RenderBSP(void *node, bool drawpsprites)
 {
+	ClearDitherTargets();
 	Bsp.Clock();
 
 	// Give the DrawInfo the viewpoint in fixed point because that's what the nodes are.
 	viewx = FLOAT2FIXED(Viewpoint.Pos.X);
 	viewy = FLOAT2FIXED(Viewpoint.Pos.Y);
+	if (r_radarclipper && !(Level->flags3 & LEVEL3_NOFOGOFWAR) && Viewpoint.IsAllowedOoB())
+	{
+		viewx = FLOAT2FIXED(Viewpoint.OffPos.X);
+		viewy = FLOAT2FIXED(Viewpoint.OffPos.Y);
+	}
 
 	validcount++;	// used for processing sidedefs only once by the renderer.
 
@@ -853,7 +1032,8 @@ void HWDrawInfo::RenderBSP(void *node, bool drawpsprites)
 		auto future = renderPool.push([&](int id) {
 			WorkerThread();
 		});
-		RenderBSPNode(node);
+		if (Viewpoint.IsOrtho() && ((Level->flags3 & LEVEL3_NOFOGOFWAR) || !r_radarclipper)) RenderOrthoNoFog();
+		else RenderBSPNode(node);
 
 		jobQueue.AddJob(RenderJob::TerminateJob, nullptr, nullptr);
 		Bsp.Unclock();
@@ -863,9 +1043,21 @@ void HWDrawInfo::RenderBSP(void *node, bool drawpsprites)
 	}
 	else
 	{
-		RenderBSPNode(node);
+		if (Viewpoint.IsOrtho() && ((Level->flags3 & LEVEL3_NOFOGOFWAR) || !r_radarclipper)) RenderOrthoNoFog();
+		else RenderBSPNode(node);
 		Bsp.Unclock();
 	}
+
+	// Make rendered targets set dither transparency flags on level geometry for next pass
+	// Can't do this inside DoSubsector() because both Trace() and P_CheckSight() affect 'validcount' global variable
+	for (int ii = 0; ii < MAXDITHERACTORS; ii++)
+	{
+		if ( RenderedTargets[ii] && P_CheckSight(players[consoleplayer].mo, RenderedTargets[ii], 0) )
+		{
+			SetDitherTransFlags(RenderedTargets[ii]);
+		}
+	}
+
 	// Process all the sprites on the current portal's back side which touch the portal.
 	if (mCurrentPortal != nullptr) mCurrentPortal->RenderAttached(this);
 
